-------------------------------------------------------------------------------
-- Title      : Programmable Clock Generator
-- Project    : 
-------------------------------------------------------------------------------
-- File       : $RCSfile: prog_clk_gen.vhd,v $
-- Author     : Cray Canada
-- Company    : Cray Canada Inc.
-- Created    : 2004-07-06
-- Last update: 2004-11-25
-------------------------------------------------------------------------------
-- Description: This block instantiates the DCMs and other circuitry to create
-- the programmable global clocks.
-------------------------------------------------------------------------------
-- Copyright (c) 2004 Cray Canada Inc.
-------------------------------------------------------------------------------
-- Revisions  :
-- $Log: prog_clk_gen.vhd,v $
-- Revision 1.1  2005/01/04 20:53:10 
-- Initial checkin.
--
-- Revision 1.2  2004/10/08 17:44:36 
-- Changed clock phase shift to ensure K clk is earlier than C clk on RAMs.
--
-- Revision 1.1  2004/08/15 17:02:16 
-- Initial checkin.
--
-------------------------------------------------------------------------------

library ieee, unisim;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use unisim.vcomponents.all;

entity prog_clock_gen is
  
  port (
    reset         : in  std_logic;
    prog_clk_in   : in  std_logic;
    k_copy_fb     : in  std_logic;
    prog_clk0     : out std_logic;
    prog_clk0_dcm : out std_logic;
    prog_clk90    : out std_logic;
    prog_clk180   : out std_logic;
    prog_dcm_lock : out std_logic);

end entity prog_clock_gen;

architecture struct of prog_clock_gen is

  -----------------------------------------------------------------------------
  -- Declare Components
  -----------------------------------------------------------------------------
  attribute box_type : string;

  component dcm
    generic (
      clk_feedback          : string  := "1X";
      clkdv_divide          : real    := 2.0;
      clkfx_divide          : integer := 1;
      clkfx_multiply        : integer := 4;
      clkin_divide_by_2     : boolean := false;   -- (true, false)
      clkin_period          : real    := 0.0;
      clkout_phase_shift    : string  := "NONE";
      deskew_adjust         : string  := "SYSTEM_SYNCHRONOUS";
      dfs_frequency_mode    : string  := "LOW";
      dll_frequency_mode    : string  := "LOW";
      dss_mode              : string  := "NONE";
      duty_cycle_correction : boolean := true;    -- (true, false)
      phase_shift           : integer := 0;
      startup_wait          : boolean := false);  -- (true, false)
    port (clk0     : out std_ulogic;
          clk180   : out std_ulogic;
          clk270   : out std_ulogic;
          clk2x    : out std_ulogic;
          clk2x180 : out std_ulogic;
          clk90    : out std_ulogic;
          clkdv    : out std_ulogic;
          clkfx    : out std_ulogic;
          clkfx180 : out std_ulogic;
          locked   : out std_ulogic;
          psdone   : out std_ulogic;
          status   : out std_logic_vector (7 downto 0);
          clkfb    : in  std_ulogic;
          clkin    : in  std_ulogic;
          dssen    : in  std_ulogic;
          psclk    : in  std_ulogic;
          psen     : in  std_ulogic;
          psincdec : in  std_ulogic;
          rst      : in  std_ulogic);
  end component;
  attribute box_type of dcm : component is "black_box";

  component clock_macro is
    generic (
      dcm_location : string);
    port (
      clock1_in  : in  std_logic;
      clock2_in  : in  std_logic;
      clock1_out : out std_logic;
      clock2_out : out std_logic);
  end component clock_macro;

  
  -----------------------------------------------------------------------------
  -- Declare Signals
  -----------------------------------------------------------------------------

  signal s_dcm1_reset_r : std_logic_vector(3 downto 0);
  signal s_dcm1_reset_delay : std_logic;
  signal s_prog_clk0_dcm_int : std_logic;
  signal s_dcm1_lock : std_logic;
  signal s_prog_clk0_inv : std_logic;
  signal s_prog_clk0_buf : std_logic;
  signal s_dcm2_reset : std_logic;
  signal s_dcm2_reset_delay : std_logic;
  signal s_prog_clk90_dcm : std_logic;
  signal s_prog_clk90_inv : std_logic;
  signal s_prog_clk90_buf : std_logic;
  
  
begin  -- architecture struct

  -----------------------------------------------------------------------------
  -- Programmable frequency clocks.
  -- The programmable clocks are generated with two chained DCMs.  The first
  -- DCM locks to the input programmable clock with 0 phase shift.  The second
  -- DCM is used to produce a clock which results in outputs with a 90 degree
  -- phase shift as seen at the IOBs.  If the clk90 output is used from the
  -- first DCM the IOB's are not shifted 90 degrees for some reason.
  -----------------------------------------------------------------------------

  -----------------------------------------------------------------------------
  -- The DCM must have a valid clock when reset to lock properly. This shift
  -- register is set when the clock is invalid (reset = '1'). When
  -- the clock becomes valid, it shifts out four reset bits through a srl16
  -- to delay them before resetting the DCM.
  prog_reset_shift : process (prog_clk_in, reset) is
  begin  -- process prog_reset_shift
    if reset = '1' then        -- asynchronous reset (active high)
      s_dcm1_reset_r <= (others => '1');
    elsif prog_clk_in'event and prog_clk_in = '1' then  -- rising clock edge
      s_dcm1_reset_r <= s_dcm1_reset_r(2 downto 0) & '0';
    end if;
  end process prog_reset_shift;

  -- Delay the reset for 16 clocks.
  prog_srl16_inst1 : srl16
    generic map (
      init => x"FFFF")
    port map (
      d   => s_dcm1_reset_r(3),
      a0  => '1',
      a1  => '1',
      a2  => '1',
      a3  => '1',
      clk => prog_clk_in,
      q   => s_dcm1_reset_delay);

  prog_clk_dcm1 : dcm
    generic map (
      clk_feedback          => "1X",
      clkin_period          => 5.0,
      clkdv_divide          => 2.0,
      clkfx_divide          => 1,
      clkfx_multiply        => 2,
      clkin_divide_by_2     => false,
      clkout_phase_shift    => "FIXED",
      deskew_adjust         => "SOURCE_SYNCHRONOUS",
      dfs_frequency_mode    => "LOW",
      dll_frequency_mode    => "LOW",
      dss_mode              => "NONE",
      duty_cycle_correction => true,
      phase_shift           => -20,
      startup_wait          => false)
    port map (
      clkin    => prog_clk_in,
      clkfb    => k_copy_fb,
      dssen    => '0',
      psincdec => '0',
      psen     => '0',
      psclk    => '0',
      rst      => s_dcm1_reset_delay,
      clk0     => s_prog_clk0_dcm_int,
      clk90    => open,
      clk180   => open,
      clk270   => open,
      clk2x    => open,
      clk2x180 => open,
      clkdv    => open,
      clkfx    => open,
      clkfx180 => open,
      locked   => s_dcm1_lock,
      psdone   => open,
      status   => open);

  prog_clk0_dcm <= s_prog_clk0_dcm_int;

  clock_macro_inst1 : clock_macro
    generic map (
      dcm_location => "x3y0")
    port map (
      clock1_in  => s_prog_clk0_dcm_int,
      clock2_in  => '0',
      clock1_out => s_prog_clk0_inv,
      clock2_out => open);

  prog_clk0_bufg : bufg
    port map (
-- i => prog_clk0_dcm,
      i => s_prog_clk0_inv,
      o => s_prog_clk0_buf);

  prog_clk0   <= s_prog_clk0_buf;
  prog_clk180 <= not s_prog_clk0_buf;
  
  s_dcm2_reset <= not s_dcm1_lock;
  
  -- Delay the reset for 16 clocks.
  prog_srl16_inst2 : srl16
    generic map (
      init => x"FFFF")
    port map (
      d   => s_dcm2_reset,
      a0  => '1',
      a1  => '1',
      a2  => '1',
      a3  => '1',
      clk => prog_clk_in,
      q   => s_dcm2_reset_delay);

  prog_clk_dcm2 : dcm
    generic map (
      clk_feedback          => "1X",
      clkin_period          => 5.0,
      clkdv_divide          => 2.0,
      clkfx_divide          => 1,
      clkfx_multiply        => 2,
      clkin_divide_by_2     => false,
      clkout_phase_shift    => "FIXED",
      deskew_adjust         => "SOURCE_SYNCHRONOUS",
      dfs_frequency_mode    => "LOW",
      dll_frequency_mode    => "LOW",
      dss_mode              => "NONE",
      duty_cycle_correction => true,
      phase_shift           => 42,
      startup_wait          => false)
    port map (
      clkin    => s_prog_clk0_buf,
      clkfb    => s_prog_clk90_buf,
      dssen    => '0',
      psincdec => '0',
      psen     => '0',
      psclk    => '0',
      rst      => s_dcm2_reset_delay,
      clk0     => s_prog_clk90_dcm,
      clk90    => open,
      clk180   => open,
      clk270   => open,
      clk2x    => open,
      clk2x180 => open,
      clkdv    => open,
      clkfx    => open,
      clkfx180 => open,
      locked   => prog_dcm_lock,
      psdone   => open,
      status   => open);

  clock_macro_inst2 : clock_macro
    generic map (
      dcm_location => "x2y0")
    port map (
      clock1_in  => s_prog_clk90_dcm,
      clock2_in  => '0',
      clock1_out => s_prog_clk90_inv,
      clock2_out => open);

  prog_clk90_bufg : bufg
    port map (
      i => s_prog_clk90_inv,
      o => s_prog_clk90_buf);

  prog_clk90 <= s_prog_clk90_buf;
  
end architecture struct;
