-------------------------------------------------------------------------------
-- FPWM Mark III
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.ALL;
use work.user_pkg.all;

entity user_app is
  port (
    -- Global signals
    reset_n       : in  std_logic;	 
    user_clk      : in  std_logic;
    user_enable   : in  std_logic;
    rt_ready      : in  std_logic;
    qdr_ready     : in  std_logic;
    -- QDR II RAM Interface
    -- RAM 1
    dr_1          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_1          : out std_logic_vector(19 downto 0);  -- read address
    aw_1          : out std_logic_vector(19 downto 0);  -- write address
    r_n_1         : out std_logic;      -- Read strobe
    w_n_1         : out std_logic;      -- Write strobe
    bw_n_1        : out std_logic_vector(7 downto 0);   -- byte write
    dw_1          : out std_logic_vector(71 downto 0);  -- write data
    -- RAM 2
    dr_2          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_2          : out std_logic_vector(19 downto 0);  -- read address
    aw_2          : out std_logic_vector(19 downto 0);  -- write address
    r_n_2         : out std_logic;      -- Read strobe
    w_n_2         : out std_logic;      -- Write strobe
    bw_n_2        : out std_logic_vector(7 downto 0);   -- byte write
    dw_2          : out std_logic_vector(71 downto 0);  -- write data
    -- RAM 3
    dr_3          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_3          : out std_logic_vector(19 downto 0);  -- read address
    aw_3          : out std_logic_vector(19 downto 0);  -- write address
    r_n_3         : out std_logic;      -- Read strobe
    w_n_3         : out std_logic;      -- Write strobe
    bw_n_3        : out std_logic_vector(7 downto 0);   -- byte write
    dw_3          : out std_logic_vector(71 downto 0);  -- write data
    -- RAM 4
    dr_4          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_4          : out std_logic_vector(19 downto 0);  -- read address
    aw_4          : out std_logic_vector(19 downto 0);  -- write address
    r_n_4         : out std_logic;      -- Read strobe
    w_n_4         : out std_logic;      -- Write strobe
    bw_n_4        : out std_logic_vector(7 downto 0);   -- byte write
    dw_4          : out std_logic_vector(71 downto 0);  -- write data
    -- RT Interface
    -- User Request Interface
    ureq_srctag   : in  std_logic_vector(4 downto 0);   -- request srctag
    ureq_full     : in  std_logic;      -- request buffer full
    ureq_notag    : in  std_logic;      -- request out of source tags
    ureq_addr     : out std_logic_vector(39 downto 3);  -- request address
    ureq_size     : out std_logic_vector(2 downto 0);   -- request size
    ureq_mask     : out std_logic_vector(7 downto 0);   -- request byte mask
    ureq_rw_n     : out std_logic;      -- request read/write
    ureq_data     : out std_logic_vector(63 downto 0);  -- request write data
    ureq_ts       : out std_logic;      -- request transfer start
    ureq_byte_req : out std_logic;      -- request type (byte or double word)
    -- Fabric Response Interface
    fresp_valid   : in  std_logic;      -- response signals valid
    fresp_ts      : in  std_logic;      -- response transfer start
    fresp_size    : in  std_logic_vector(2 downto 0);   -- response size
    fresp_srctag  : in  std_logic_vector(4 downto 0);   -- response srctag
    fresp_data    : in  std_logic_vector(63 downto 0);  -- response data
    fresp_enable  : out std_logic;      -- enable responses
    -- Fabric Request Interface
    freq_addr     : in  std_logic_vector(39 downto 3);  -- request address
    freq_size     : in  std_logic_vector(3 downto 0);   -- request size
    freq_srctag   : in  std_logic_vector(4 downto 0);   -- request srctag
    freq_mask     : in  std_logic_vector(7 downto 0);   -- request byte mask
    freq_rw_n     : in  std_logic;      -- request read/write 
    freq_ts       : in  std_logic;      -- request transfer start
    freq_valid    : in  std_logic;      -- request valid
    freq_data     : in  std_logic_vector(63 downto 0);  -- request write data
    freq_enable   : out std_logic;      -- enable request interface
    -- User Response Interface
    uresp_full    : in  std_logic;      -- response buffer full
    uresp_ts      : out std_logic;      -- response transfer start
    uresp_size    : out std_logic_vector(3 downto 0);   -- response size
    uresp_srctag  : out std_logic_vector(4 downto 0);   -- response srctag
    uresp_data    : out std_logic_vector(63 downto 0)   -- response data
    );

end entity user_app;


architecture rtl of user_app is
	signal ftr_pointer : std_logic_vector(39 downto 3);
	
	signal roundcount_gen, roundcount_write: std_logic_vector(3 downto 0);
	signal start_count, count_gen, count_write: std_logic_vector(15 downto 0);

	signal sram_addr : std_logic_vector(19 downto 0);

	signal writing, warmedup, stall: std_logic;

	signal burst_size : std_logic_vector(3 downto 0);
	signal burst_addr : std_logic_vector(26 downto 3);

	signal pwm_data_in : std_logic_vector(127 downto 0);
	signal pwm_data_out: std_logic_vector(37 downto 0);
	signal pwm_index_in, pwm_index_out: std_logic_vector(26 downto 0);
	signal pwm_done, pwm_valid, pwm_data_out_valid : std_logic;

	signal write_addr : std_logic_vector(39 downto 3);
	signal write_data : std_logic_vector(63 downto 0);
	signal write_valid, write_bypass, write_flush, write_set_base_addr, write_empty, write_overflow : std_logic;

	component pwm is
	 port(
	   reset_n     : in  std_logic;
		user_clk		: in  std_logic;
		stall			: in  std_logic;
		valid			: in  std_logic;
		done			: in  std_logic;

		-- Input lines
		data_in		: in  std_logic_vector(63 downto 0);
		index_in		: in  std_logic_vector(26 downto 0);

		-- Output lines
		data_out			: out std_logic_vector(37 downto 0);
		index_out		: out std_logic_vector(26 downto 0);
		data_out_valid : out std_logic
	 );
	end component pwm;

	component writer is
	 port (
		reset_n			: in  std_logic;	 
		user_clk			: in  std_logic;
		valid				: in  std_logic;
		bypass			: in  std_logic;
		flush				: in  std_logic;
		set_base_addr	: in  std_logic;

		data_in			: in  std_logic_vector(63 downto 0);
		addr_in			: in  std_logic_vector(39 downto 3);

		stall				: out std_logic;
		empty				: out std_logic;
		overflow			: out std_logic;

		ureq_srctag   	: in  std_logic_vector(4 downto 0); -- unused
		ureq_full		: in  std_logic;
		ureq_notag		: in  std_logic; -- unused
		ureq_addr		: out std_logic_vector(39 downto 3);
		ureq_size		: out std_logic_vector(2 downto 0);
		ureq_mask		: out std_logic_vector(7 downto 0);
		ureq_rw_n		: out std_logic;
		ureq_data		: out std_logic_vector(63 downto 0);
		ureq_ts			: out std_logic;
		ureq_byte_req	: out std_logic
	 );
	end component pwm;

begin  -- architecture rtl

	-- Stub unused parts of SRAM interface (byte filter)
	bw_n_1        <= (others => '0');
   bw_n_2        <= (others => '0');
   bw_n_3        <= (others => '0');
   bw_n_4        <= (others => '0');

	-- Stub unused parts of RT interface (fabric response -- no reads issued from FPGA)
   fresp_enable  <= '0';

	rt_xfer : process (reset_n, user_clk)
	begin
		if reset_n='0' then
			ftr_pointer <= (others => '0');
			start_count <= (others => '0');

		   aw_1          <= (others => '0');
		   w_n_1         <= '1';
		   dw_1          <= (others => '0');

		   aw_2          <= (others => '0');
		   w_n_2         <= '1';
		   dw_2          <= (others => '0');

		   aw_3          <= (others => '0');
		   w_n_3         <= '1';
		   dw_3          <= (others => '0');

		   aw_4          <= (others => '0');
		   w_n_4         <= '1';
		   dw_4          <= (others => '0');

 			uresp_ts      <= '0';
			uresp_size    <= (others => '0');
			uresp_srctag  <= (others => '0');
			uresp_data    <= (others => '0');

			freq_enable   <= '0';

			burst_size <= (others => '0');
			burst_addr <= (others => '0');

			writing <= '1';
		elsif user_clk='1' and user_clk'event then
			-- SMP-side reads and writes
	  		if rt_ready='1' and qdr_ready='1' and user_enable='1' then
				-- write request
				if burst_size > X"0" then
					burst_size <= burst_size - 1;
					burst_addr <= burst_addr + 1;

					case burst_addr(24 downto 23) is 
						when "00" =>
							aw_1 <= burst_addr(22 downto 3);
							dw_1 <= insert_parity(freq_data);
							w_n_1 <= '0';
						when "01" =>
							aw_2 <= burst_addr(22 downto 3);
							dw_2 <= insert_parity(freq_data);
							w_n_2 <= '0';			
						when "10" =>
							aw_3 <= burst_addr(22 downto 3);
							dw_3 <= insert_parity(freq_data);
							w_n_3 <= '0';
						when "11" =>	
							aw_4 <= burst_addr(22 downto 3);
							dw_4 <= insert_parity(freq_data);
							w_n_4 <= '0';
						when others =>
							null;
					end case;

					uresp_data <= freq_data; -- debug
				elsif writing='1' then
					freq_enable <= '1';
					uresp_ts <= '0';
					uresp_srctag <= (others => '0');
					uresp_data <= (others => '0');
					writing <= '0';					
				elsif freq_valid='1' and freq_ts='1' then
					-- the start of some kind of access...
					if freq_rw_n='1' then
						-- .. which is a read.. 
						case freq_addr(22 downto 3) is
							when X"0000A" =>
								uresp_data(39 downto 3) <= ftr_pointer;
							when X"0000B" =>
								uresp_data(15 downto 0) <= count_gen;
							when X"0000C" =>
								uresp_data(15 downto 0) <= start_count;
							when others => 
								uresp_data(26 downto 3) <= freq_addr(26 downto 3);
								uresp_data(63 downto 27) <= (others => '1');
						end case;

						uresp_ts <= '1';
						uresp_srctag <= freq_srctag;
						freq_enable <= '0';
						writing <= '1';
					elsif freq_rw_n='0' then
						-- .. which is a write...
						if freq_addr(25)='1' then
							-- .. to the SRAM..
							if freq_size>X"0" then
								-- .. and will continue as a burst.
								burst_size <= freq_size - 1;
								burst_addr <= freq_addr(26 downto 3) + 1;
							end if;

							-- either way, read and store this quadword.
							case freq_addr(24 downto 23) is 
								when "00" =>
									aw_1 <= freq_addr(22 downto 3);
									dw_1 <= insert_parity(freq_data);
									w_n_1 <= '0';
								when "01" =>
									aw_2 <= freq_addr(22 downto 3);
									dw_2 <= insert_parity(freq_data);
									w_n_2 <= '0';			
								when "10" =>
									aw_3 <= freq_addr(22 downto 3);
									dw_3 <= insert_parity(freq_data);
									w_n_3 <= '0';
								when "11" =>	
									aw_4 <= freq_addr(22 downto 3);
									dw_4 <= insert_parity(freq_data);
									w_n_4 <= '0';
								when others =>
									null;
							end case;

							uresp_data <= freq_data; -- debug
						elsif freq_addr(25)='0' then
							-- .. to a variable (these are assumed to be set using fpga_wrt_appif_val
							case freq_addr(22 downto 3) is
								when X"0000A" =>
									-- sets writeback address (FTR)
									ftr_pointer <= freq_data(39 downto 3);
									uresp_data(39 downto 3) <= freq_data(39 downto 3); -- debug
								when X"0000B" =>
									-- sets number of cycles to run (8 quadword results [64 bytes] per round)
									-- FTR mem *must* hold the required amount of data, or data corruption may occur!
									-- THERE IS NO MEMORY PROTECTION, AND CHECKING SMP'S BUFFER FROM FPGA IS IMPOSSIBLE!
									start_count <= freq_data(15 downto 0);
									uresp_data(15 downto 0) <= freq_data(15 downto 0); -- debug
								when others =>
									null;
							end case;
						end if;
					end if;
				else
					start_count <= (others => '0');

					w_n_1 <= '1';
					w_n_2 <= '1';
					w_n_3 <= '1';
					w_n_4 <= '1';
				end if;
			end if;
		end if;
	end process rt_xfer;



	data_push : process (reset_n, user_clk)
	begin	
		if reset_n='0' then
			-- sets initial values
			sram_addr <= (others => '0');

			ar_1 <= (others => '0');
			ar_2 <= (others => '0');
			ar_3 <= (others => '0');
			ar_4 <= (others => '0');
			
			r_n_1 <= '1';
			r_n_2 <= '1';
			r_n_3 <= '1';
			r_n_4 <= '1';

			roundcount_gen <= (others => '0');
			count_gen <= (others => '0');

			pwm_data_in <= (others => '1');
			pwm_index_in <= (others => '1');

			pwm_done <= '0';
			pwm_valid <= '0';
			
			stall <= '0';
			warmedup <= '0';

			write_valid <= '0';
			write_bypass <= '0';
			write_flush <= '0';
			write_set_base_addr <= '0';
		elsif user_clk='1' and user_clk'event then
	  		if rt_ready='1' and qdr_ready='1' and user_enable='1' then
				if count_gen > 0 and ftr_pointer /= 0 then
					write_set_base_addr <= '0';
										
					if roundcount_gen = X"8" then
						roundcount_gen <= (others => '0');

						ar_1 <= sram_addr;
						ar_2 <= sram_addr;
						ar_3 <= sram_addr;
						ar_4 <= sram_addr;

						r_n_1 <= '0';
						r_n_2 <= '0';
						r_n_3 <= '0';
						r_n_4 <= '0';

						sram_addr <= sram_addr + 1;

						warmedup <= '1';

						if warmedup='1' then
							--pwm_data_in(63 downto 0) <= pwm_data_in(71 downto 8);
							--pwm_data_in(127 downto 64) <= extract_data(dr_1);
							--pwm_index_in(3 downto 0) <= roundcount_gen; -- fixme: real index
							count_gen <= count_gen-1;
							pwm_data_in(63 downto 0) <= 	"0000" & roundcount_gen + 0 mod 4 &
																	"0000" & roundcount_gen + 1 mod 4 &
																	"0000" & roundcount_gen + 2 mod 4 &
																	"0000" & roundcount_gen + 3 mod 4 &
																	"0000" & roundcount_gen + 0 mod 4 &
																	"0000" & roundcount_gen + 1 mod 4 &
																	"0000" & roundcount_gen + 2 mod 4 &
																	"0000" & roundcount_gen + 3 mod 4;
							pwm_valid <= '1';
						end if;						
					else
						pwm_data_in(119 downto 0) <= pwm_data_in(127 downto 8);
						pwm_data_in(127 downto 120) <= (others => '0'); 
						roundcount_gen <= roundcount_gen + 1;
					end if;
				else
					count_gen <= start_count;
					roundcount_gen <= X"8";
					sram_addr <= (others => '0');
					write_addr <= ftr_pointer;
					write_set_base_addr <= '1';

					ar_1 <= (others => '0');
					ar_2 <= (others => '0');
					ar_3 <= (others => '0');
					ar_4 <= (others => '0');

					r_n_1 <= '1';
					r_n_2 <= '1';
					r_n_3 <= '1';
					r_n_4 <= '1';

					pwm_valid <= '0';
				end if;

				if pwm_data_out_valid='1' then
					write_data <= pwm_index_out & pwm_data_out;
					write_valid <= '1';
				else
					write_data <= (others => '0');
					write_valid <= '0';
				end if;
			end if;
		end if;
	end process data_push;


	pwm_inst : pwm
	port map(
		-- Global control signal
		reset_n => reset_n,
		user_clk => user_clk,
		valid => pwm_valid,
		stall => stall,
		done => pwm_done,

		-- Input signals
		data_in => pwm_data_in(63 downto 0),
		index_in => pwm_index_in,

		-- Output signals
		data_out => pwm_data_out,
		index_out => pwm_index_out,
		data_out_valid => pwm_data_out_valid
	);

	-- fixme: add filter/sort module here

	write_inst : writer
	port map(
		reset_n => reset_n,
		user_clk => user_clk,
		valid => write_valid,
		bypass => write_bypass,
		flush => write_flush,
		set_base_addr => write_set_base_addr,

		data_in => write_data,
		addr_in => write_addr,

		stall => write_stall,
		empty => write_empty,
		overflow => write_overflow,

		ureq_srctag => ureq_srctag,
		ureq_full => ureq_full,
		ureq_notag => ureq_notag,
		ureq_addr => ureq_addr,
		ureq_size => ureq_size,
		ureq_mask => ureq_mask,
		ureq_rw_n => ureq_rw_n,
		ureq_data => ureq_data,
		ureq_ts => ureq_ts,
		ureq_byte_req => ureq_byte_req
	);

end architecture rtl;
