-------------------------------------------------------------------------------
-- FPWM Mark III
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.ALL;
use work.user_pkg.all;

entity user_app is
  port (
	debug : out std_logic_vector(25 downto 0);

    -- Global signals
    reset_n       : in  std_logic;	 
    user_clk      : in  std_logic;
    user_enable   : in  std_logic;
    rt_ready      : in  std_logic;
    qdr_ready     : in  std_logic;
    -- QDR II RAM Interface
    -- RAM 1
    dr_1          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_1          : out std_logic_vector(19 downto 0);  -- read address
    aw_1          : out std_logic_vector(19 downto 0);  -- write address
    r_n_1         : out std_logic;      -- Read strobe
    w_n_1         : out std_logic;      -- Write strobe
    bw_n_1        : out std_logic_vector(7 downto 0);   -- byte write
    dw_1          : out std_logic_vector(71 downto 0);  -- write data
    -- RAM 2
    dr_2          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_2          : out std_logic_vector(19 downto 0);  -- read address
    aw_2          : out std_logic_vector(19 downto 0);  -- write address
    r_n_2         : out std_logic;      -- Read strobe
    w_n_2         : out std_logic;      -- Write strobe
    bw_n_2        : out std_logic_vector(7 downto 0);   -- byte write
    dw_2          : out std_logic_vector(71 downto 0);  -- write data
    -- RAM 3
    dr_3          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_3          : out std_logic_vector(19 downto 0);  -- read address
    aw_3          : out std_logic_vector(19 downto 0);  -- write address
    r_n_3         : out std_logic;      -- Read strobe
    w_n_3         : out std_logic;      -- Write strobe
    bw_n_3        : out std_logic_vector(7 downto 0);   -- byte write
    dw_3          : out std_logic_vector(71 downto 0);  -- write data
    -- RAM 4
    dr_4          : in  std_logic_vector(71 downto 0);  -- read data 
    ar_4          : out std_logic_vector(19 downto 0);  -- read address
    aw_4          : out std_logic_vector(19 downto 0);  -- write address
    r_n_4         : out std_logic;      -- Read strobe
    w_n_4         : out std_logic;      -- Write strobe
    bw_n_4        : out std_logic_vector(7 downto 0);   -- byte write
    dw_4          : out std_logic_vector(71 downto 0);  -- write data
    -- RT Interface
    -- User Request Interface
    ureq_srctag   : in  std_logic_vector(4 downto 0);   -- request srctag
    ureq_full     : in  std_logic;      -- request buffer full
    ureq_notag    : in  std_logic;      -- request out of source tags
    ureq_addr     : out std_logic_vector(39 downto 3);  -- request address
    ureq_size     : out std_logic_vector(2 downto 0);   -- request size
    ureq_mask     : out std_logic_vector(7 downto 0);   -- request byte mask
    ureq_rw_n     : out std_logic;      -- request read/write
    ureq_data     : out std_logic_vector(63 downto 0);  -- request write data
    ureq_ts       : out std_logic;      -- request transfer start
    ureq_byte_req : out std_logic;      -- request type (byte or double word)
    -- Fabric Response Interface
    fresp_valid   : in  std_logic;      -- response signals valid
    fresp_ts      : in  std_logic;      -- response transfer start
    fresp_size    : in  std_logic_vector(2 downto 0);   -- response size
    fresp_srctag  : in  std_logic_vector(4 downto 0);   -- response srctag
    fresp_data    : in  std_logic_vector(63 downto 0);  -- response data
    fresp_enable  : out std_logic;      -- enable responses
    -- Fabric Request Interface
    freq_addr     : in  std_logic_vector(39 downto 3);  -- request address
    freq_size     : in  std_logic_vector(3 downto 0);   -- request size
    freq_srctag   : in  std_logic_vector(4 downto 0);   -- request srctag
    freq_mask     : in  std_logic_vector(7 downto 0);   -- request byte mask
    freq_rw_n     : in  std_logic;      -- request read/write 
    freq_ts       : in  std_logic;      -- request transfer start
    freq_valid    : in  std_logic;      -- request valid
    freq_data     : in  std_logic_vector(63 downto 0);  -- request write data
    freq_enable   : out std_logic;      -- enable request interface
    -- User Response Interface
    uresp_full    : in  std_logic;      -- response buffer full
    uresp_ts      : out std_logic;      -- response transfer start
    uresp_size    : out std_logic_vector(3 downto 0);   -- response size
    uresp_srctag  : out std_logic_vector(4 downto 0);   -- response srctag
    uresp_data    : out std_logic_vector(63 downto 0)   -- response data
    );

end entity user_app;


architecture rtl of user_app is
	signal ftr_pointer : std_logic_vector(39 downto 3);
	signal ftr_pointer_is_set : std_logic; -- separated as a signal due to difficulties meeting timing 

	signal start_count, count_gen: std_logic_vector(23 downto 0);
	signal start_count_is_set : std_logic;
	
	signal index_counter : std_logic_vector(25 downto 0);
	signal count_cooldown : std_logic_vector(7 downto 0);

	signal writing, stall: std_logic;

	signal burst_size : std_logic_vector(3 downto 0);
	signal burst_addr : std_logic_vector(24 downto 3);

	signal pwm_elem_in : std_logic_vector(7 downto 0);
	signal pwm_data_out: std_logic_vector(37 downto 0);
	signal pwm_index_in, pwm_index_out: std_logic_vector(25 downto 0);
	signal pwm_valid, pwm_data_out_valid : std_logic;

	signal write_addr : std_logic_vector(39 downto 3);
	signal write_data : std_logic_vector(63 downto 0);
	signal write_valid, write_bypass, write_flush, write_set_base_addr, write_empty, write_overflow : std_logic;

	signal pwm_table: lu_table;

	signal reader_active, reader_out_valid : std_logic;
	signal reader_data_offset : std_logic_vector(25 downto 3);
	signal reader_data_out : std_logic_vector(7 downto 0);
	signal reader_index_out : std_logic_vector(25 downto 0);

	signal sram_write_data : std_logic_vector(63 downto 0);
	signal sram_write_addr : std_logic_vector(22 downto 3);
	signal sram_write_sel : std_logic_vector(1 downto 0);
	signal sram_write_valid : std_logic;

	signal pwm_write_data : std_logic_vector(31 downto 0);
	signal pwm_write_row : std_logic_vector(6 downto 0);
	signal pwm_write_col : std_logic_vector(13 downto 0);
	signal pwm_write_valid : std_logic;

	-- Added signals for filter
	signal filter_data_in : std_logic_vector(37 downto 0);
	signal filter_data_out : std_logic_vector(63 downto 0);
	signal filter_index_in : std_logic_vector(25 downto 0);
	signal filter_valid_in : std_logic;
	signal filter_valid_out : std_logic;

	signal thresh : std_logic_vector(37 downto 0);

	-- timing is fun
   signal tmp_freq_data     : std_logic_vector(63 downto 0);
   signal tmp_freq_addr     : std_logic_vector(39 downto 3);
   signal tmp_freq_size     : std_logic_vector(3 downto 0);
   signal tmp_freq_srctag   : std_logic_vector(4 downto 0);
   signal tmp_freq_mask     : std_logic_vector(7 downto 0);
   signal tmp_freq_rw_n     : std_logic;
   signal tmp_freq_ts       : std_logic;
   signal tmp_freq_valid    : std_logic;

	component reader is
	 port(
		reset_n     : in  std_logic;
		user_clk		: in  std_logic;
		stall			: in  std_logic;

		active		: in  std_logic;
		data_offset	: in  std_logic_vector(25 downto 3);

		out_valid	: out std_logic;
		data_out		: out std_logic_vector(7 downto 0);
		index_out	: out std_logic_vector(25 downto 0);

		-- sram interface
		dr_1        : in  std_logic_vector(71 downto 0);  -- read data 
		ar_1        : out std_logic_vector(19 downto 0);  -- read address
		r_n_1       : out std_logic;      -- Read strobe

		dr_2        : in  std_logic_vector(71 downto 0);  -- read data 
		ar_2        : out std_logic_vector(19 downto 0);  -- read address
		r_n_2       : out std_logic;      -- Read strobe

		dr_3        : in  std_logic_vector(71 downto 0);  -- read data 
		ar_3        : out std_logic_vector(19 downto 0);  -- read address
		r_n_3       : out std_logic;      -- Read strobe

		dr_4        : in  std_logic_vector(71 downto 0);  -- read data 
		ar_4        : out std_logic_vector(19 downto 0);  -- read address
		r_n_4       : out std_logic      -- Read strobe
	 );
	end component reader;


	component pwm is
	 port(
	   reset_n     : in  std_logic;
		user_clk		: in  std_logic;
		stall			: in  std_logic;
		valid			: in  std_logic;

		-- Input lines
		elem_in		: in  std_logic_vector(7 downto 0);
		index_in		: in  std_logic_vector(25 downto 0);
		pwm_table 	: in  lu_table;

		-- Output lines
		data_out			: out std_logic_vector(37 downto 0);
		index_out		: out std_logic_vector(25 downto 0);
		data_out_valid : out std_logic
	 );
	end component pwm;

	component filter is
	 port (
		reset_n 		: in  std_logic;
		user_clk 		: in  std_logic;
		valid 			: in  std_logic;
		stall			: in  std_logic;
		data_in 		: in  std_logic_vector(37 downto 0);
		index_in		: in  std_logic_vector(25 downto 0);
		threshold		: in  std_logic_vector(37 downto 0);

		data_valid		: out std_logic;
		data_out		: out std_logic_vector(63 downto 0)
	 );
	end component filter;

	component writer is
	 port (
		reset_n			: in  std_logic;	 
		user_clk			: in  std_logic;
		valid				: in  std_logic;
		bypass			: in  std_logic;
		flush				: in  std_logic;
		set_base_addr	: in  std_logic;

		data_in			: in  std_logic_vector(63 downto 0);
		addr_in			: in  std_logic_vector(39 downto 3);

		stall				: out std_logic;
		empty				: out std_logic;
		overflow			: out std_logic;

		ureq_srctag   	: in  std_logic_vector(4 downto 0); -- unused
		ureq_full		: in  std_logic;
		ureq_notag		: in  std_logic; -- unused
		ureq_addr		: out std_logic_vector(39 downto 3);
		ureq_size		: out std_logic_vector(2 downto 0);
		ureq_mask		: out std_logic_vector(7 downto 0);
		ureq_rw_n		: out std_logic;
		ureq_data		: out std_logic_vector(63 downto 0);
		ureq_ts			: out std_logic;
		ureq_byte_req	: out std_logic
	 );
	end component writer;

begin  -- architecture rtl

	-- Stub unused parts of SRAM interface (byte filter)
	bw_n_1        <= (others => '0');
   bw_n_2        <= (others => '0');
   bw_n_3        <= (others => '0');
   bw_n_4        <= (others => '0');

	-- Stub unused parts of RT interface (fabric response -- no reads issued from FPGA)
   fresp_enable  <= '0';

	rt_xfer : process (reset_n, user_clk)
	begin
		if reset_n='0' then
			ftr_pointer <= (others => '0');
			ftr_pointer_is_set <= '0';

			start_count <= (others => '0');
			start_count_is_set <= '0';
										
 			uresp_ts      <= '0';
			uresp_size    <= (others => '0');
			uresp_srctag  <= (others => '0');
			uresp_data    <= (others => '0');

			freq_enable   <= '0';

			sram_write_valid <= '0';
			sram_write_data <= (others => '0');
			sram_write_addr <= (others => '0');
			sram_write_sel <= (others => '0');
						
			burst_size <= (others => '0');
			burst_addr <= (others => '0');

			writing <= '1';
			
			thresh <= (others => '0');

			tmp_freq_data <= (others => '0');	
			tmp_freq_addr <= (others => '0');
			tmp_freq_size <= (others => '0');
			tmp_freq_srctag <= (others => '0');
			tmp_freq_mask <= (others => '0');
			tmp_freq_rw_n <= '0';
			tmp_freq_ts <= '0';
			tmp_freq_valid <= '0';
		elsif user_clk='1' and user_clk'event then
			-- SMP-side reads and writes

			tmp_freq_data <= freq_data;	
			tmp_freq_addr <= freq_addr;
			tmp_freq_size <= freq_size;
			tmp_freq_srctag <= freq_srctag;
			tmp_freq_mask <= freq_mask;
			tmp_freq_rw_n <= freq_rw_n;
			tmp_freq_ts <= freq_ts;
			tmp_freq_valid <= freq_valid;


			if user_enable='1' then
				-- determine sram valid signal
				if or_reduce(burst_size)='1' or (tmp_freq_valid='1' and tmp_freq_ts='1' and tmp_freq_rw_n='0' and tmp_freq_addr(25)='1') then
					sram_write_valid <= '1';
				else
					sram_write_valid <= '0';
				end if;

				-- determine pwm valid signal
				if tmp_freq_valid='1' and tmp_freq_ts='1' and tmp_freq_rw_n='0' and tmp_freq_addr(25)='0' and tmp_freq_addr(24)='1' then
					pwm_write_valid <= '1';
				else
					pwm_write_valid <= '0';
				end if;

				-- write request
				if or_reduce(burst_size)='1' then
					burst_size <= burst_size - 1;
					burst_addr <= burst_addr + 1;

					sram_write_addr <= burst_addr(22 downto 3);
					sram_write_sel <= burst_addr(24 downto 23);
					sram_write_data <= tmp_freq_data;
				elsif writing='1' then
					freq_enable <= '1';
					uresp_ts <= '0';
					uresp_srctag <= (others => '0');
					uresp_data <= (others => '0');
					writing <= '0';					
				elsif tmp_freq_valid='1' and tmp_freq_ts='1' then
					-- the start of some kind of access...
					if tmp_freq_rw_n='1' then
						-- .. which is a read.. 
						case tmp_freq_addr(6 downto 3) is
							when X"A" =>
								uresp_data(39 downto 3) <= ftr_pointer;
							-- perandg's add for threshold transfer
							when X"4" =>
								uresp_data(37 downto 0) <= thresh;
							when X"F" =>
								uresp_data(23 downto 0) <= start_count;
							when others => 
								uresp_data(26 downto 3) <= tmp_freq_addr(26 downto 3);
								uresp_data(63 downto 27) <= (others => '1');
						end case;

						uresp_ts <= '1';
						uresp_srctag <= tmp_freq_srctag;
						freq_enable <= '0';
						writing <= '1';
					elsif tmp_freq_rw_n='0' then
						-- .. which is a write...
						if tmp_freq_addr(25)='1' then
							-- .. to the SRAM..
							if tmp_freq_size>X"0" then
								-- .. and will continue as a burst.
								burst_size <= tmp_freq_size - 1;
								burst_addr <= tmp_freq_addr(24 downto 3) + 1;
							end if;

							-- either way, read and store this quadword.
							sram_write_addr <= tmp_freq_addr(22 downto 3);
							sram_write_sel <= tmp_freq_addr(24 downto 23);
							sram_write_data <= tmp_freq_data;
						elsif tmp_freq_addr(24)='1' then
							-- .. to the PWM lookup table (warning: there is no burst support for this)
							pwm_write_data <=	tmp_freq_data(31 downto 0);
							pwm_write_col <= tmp_freq_addr(23 downto 10);
							pwm_write_row <= tmp_freq_addr(9 downto 3); 
						else
							-- .. to a variable (these are assumed to be set using fpga_wrt_appif_val
							case tmp_freq_addr(6 downto 3) is
								when X"A" =>
									-- sets writeback address (FTR)
									report "set ftr_pointer";
									ftr_pointer <= tmp_freq_data(39 downto 3);
									ftr_pointer_is_set <= '1';
								when X"4" =>
									---Sets threshold value for filter
									report "set threshold value";
									thresh <= tmp_freq_data(37 downto 0);
									-- thresh <= "00000000000000000000000000000000000101";
								when X"F" =>
									-- sets number of cycles to run
									-- FTR mem *must* hold the required amount of data, or data corruption may occur!
									-- THERE IS NO MEMORY PROTECTION, AND CHECKING SMP'S BUFFER FROM FPGA IS IMPOSSIBLE!
									report "set start_count";
									start_count <= tmp_freq_data(23 downto 0);
									start_count_is_set <= '1';
								when others =>
									null;
							end case;
						end if;
					end if;
				else
					start_count <= (others => '0');
					start_count_is_set <= '0';
				end if;
			end if;
		end if;
	end process rt_xfer;

	sram_write : process (reset_n, user_clk)
	begin
		if reset_n='0' then
		   aw_1          <= (others => '0');
		   w_n_1         <= '1';
		   dw_1          <= (others => '0');

		   aw_2          <= (others => '0');
		   w_n_2         <= '1';
		   dw_2          <= (others => '0');

		   aw_3          <= (others => '0');
		   w_n_3         <= '1';
		   dw_3          <= (others => '0');

		   aw_4          <= (others => '0');
		   w_n_4         <= '1';
		   dw_4          <= (others => '0');		
		elsif user_clk='1' and user_clk'event then
			-- sram write stuff
			if qdr_ready='1' then 
				if sram_write_valid='1' then
					if sram_write_sel="00" then
						aw_1 <= sram_write_addr;
						dw_1 <= insert_parity(sram_write_data);
						w_n_1 <= '0';
					else
						w_n_1 <= '1';
					end if;

					if sram_write_sel="01" then
						aw_2 <= sram_write_addr;
						dw_2 <= insert_parity(sram_write_data);
						w_n_2 <= '0';
					else
						w_n_2 <= '1';
					end if;

					if sram_write_sel="10" then
						aw_3 <= sram_write_addr;
						dw_3 <= insert_parity(sram_write_data);
						w_n_3 <= '0';
					else
						w_n_3 <= '1';
					end if;

					if sram_write_sel="11" then
						aw_4 <= sram_write_addr;
						dw_4 <= insert_parity(sram_write_data);
						w_n_4 <= '0';
					else
						w_n_4 <= '1';
					end if;
				else
					w_n_1 <= '1';
					w_n_2 <= '1';
					w_n_3 <= '1';
					w_n_4 <= '1';
				end if;
			end if;
		end if;
	end process sram_write;
	
	pwm_write : process (reset_n, user_clk)
	begin
		if reset_n='0' then
			pwm_table <= (others => (others => '0'));		
		elsif user_clk='1' and user_clk'event then
			if pwm_write_valid='1' then
				pwm_set(pwm_write_col(2 downto 0), pwm_write_row, pwm_write_data, pwm_table);
			end if;
		end if;
	end process pwm_write;
	

	data_push : process (reset_n, user_clk)
	begin	
		if reset_n='0' then
			-- sets initial values
			count_gen <= (others => '0');
			count_cooldown <= (others => '0');
			index_counter <= (others => '0');

			pwm_elem_in <= (others => '1');
			pwm_index_in <= (others => '1');

			pwm_valid <= '0';
			write_data <= (others => '0');
			write_valid <= '0';
			write_bypass <= '0';
			write_flush <= '0';
			write_set_base_addr <= '0';

			filter_data_in <= (others => '0');
			filter_index_in <= (others => '0');
			filter_valid_in <= '0';

			debug <= (others => '0');
		elsif user_clk='1' and user_clk'event then
	  		if user_enable='1' then
				if or_reduce(count_gen)='1' and ftr_pointer_is_set='1' then	-- count_gen > 0 resulted in a carry chain with too much delay to work
					reader_active <= '1';
					write_set_base_addr <= '0';
					write_flush <= '0';
					count_cooldown <= X"10";

					if stall = '0' then
						count_gen <= count_gen-1;
					end if;
					
				elsif count_cooldown > 0 then
					pwm_valid <= '0';
					count_cooldown <= count_cooldown-1;
					debug(25 downto 8) <= (others => '1');
					debug(7 downto 0) <= count_cooldown;
				else
					if start_count_is_set='1' then
						reader_data_offset <= (others => '0'); -- fixme: make offset-able
						count_gen <= start_count;
					else
						reader_data_offset <= (others => '0');
						count_gen <= (others => '0');
					end if;
				
					reader_active <= '0';
					write_addr <= ftr_pointer;
					write_set_base_addr <= '1';
					pwm_valid <= '0';
					write_flush <= '1';
					debug <= (others => '0');
				end if;
			end if;
		end if;
	end process data_push;

	reader_inst : reader
	port map(
		-- Global control signal
		reset_n => reset_n,
		user_clk => user_clk,
		stall => stall,

		active		=>	reader_active,
		data_offset	=> reader_data_offset,

		out_valid	=> reader_out_valid,
		data_out		=> reader_data_out,
		index_out	=> reader_index_out,

		dr_1  => dr_1,
		ar_1  => ar_1,
		r_n_1 =>	r_n_1,

		dr_2  => dr_2,
		ar_2  =>	ar_2,
		r_n_2 =>	r_n_2,

		dr_3  => dr_3,
		ar_3  => ar_3,
		r_n_3 =>	r_n_3,

		dr_4  => dr_4,
		ar_4  =>	ar_4,
		r_n_4 =>	r_n_4
	);

	pwm_inst : pwm
	port map(
		-- Global control signal
		reset_n => reset_n,
		user_clk => user_clk,
		valid => reader_out_valid,
		stall => stall,

		-- Input signals
		elem_in => reader_data_out,
		index_in => reader_index_out,
		pwm_table => pwm_table,

		-- Output signals
		data_out => pwm_data_out,
		index_out => pwm_index_out,
		data_out_valid => pwm_data_out_valid
	);

	filter_inst : filter
	port map(
		reset_n => reset_n,
		user_clk => user_clk,
		valid => pwm_data_out_valid,
		stall => stall,
		data_in => pwm_data_out,
		index_in => pwm_index_out,
		threshold => thresh,
  
		data_valid => filter_valid_out,
		data_out => filter_data_out
	);

	write_inst : writer
	port map(
		reset_n => reset_n,
		user_clk => user_clk,
		valid => filter_valid_out,
		bypass => write_bypass,
		flush => write_flush,
		set_base_addr => write_set_base_addr,

		data_in => filter_data_out,
		addr_in => write_addr,

		stall => stall,
		empty => write_empty,
		overflow => write_overflow,

		ureq_srctag => ureq_srctag,
		ureq_full => ureq_full,
		ureq_notag => ureq_notag,
		ureq_addr => ureq_addr,
		ureq_size => ureq_size,
		ureq_mask => ureq_mask,
		ureq_rw_n => ureq_rw_n,
		ureq_data => ureq_data,
		ureq_ts => ureq_ts,
		ureq_byte_req => ureq_byte_req
	);

end architecture rtl;
