--------------------------------------------------------------------------------
-- todo: make bypass functionality
-- todo: make sticky flush
-- note: after flush is flagged, no writes should be issued before the writer flags "empty".
-- note: base address setting should ONLY be attempted when the circuit is quiescent (empty, no other in flags)	
--------------------------------------------------------------------------------

-----------------------
-- Write and read means read from BRAM (and write to CPU) and write to BRAM.
-----------------------

library ieee, unisim;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use work.user_pkg.all;
use unisim.VCOMPONENTS.all;

entity writer is
	port (
		reset_n				: in  std_logic;	 
		user_clk				: in  std_logic;
		valid					: in  std_logic;
		bypass				: in  std_logic;
		flush					: in  std_logic;
		set_base_addr		: in  std_logic;

		data_in				: in  std_logic_vector(63 downto 0);
		addr_in				: in  std_logic_vector(39 downto 3);

		stall					: out std_logic;
		empty					: out std_logic;
		overflow				: out std_logic;

		ureq_srctag   		: in  std_logic_vector(4 downto 0); -- unused
		ureq_full			: in  std_logic;
		ureq_notag			: in  std_logic; -- unused
		ureq_addr			: out std_logic_vector(39 downto 3);
		ureq_size			: out std_logic_vector(2 downto 0);
		ureq_mask			: out std_logic_vector(7 downto 0);
		ureq_rw_n			: out std_logic;
		ureq_data			: out std_logic_vector(63 downto 0);
		ureq_ts				: out std_logic;
		ureq_byte_req		: out std_logic
	);
end writer;

architecture rtl of writer is
	component RAMB4_S16_S16
        port (
                ADDRA:  in std_logic_vector(7 downto 0);
                ADDRB:  in std_logic_vector(7 downto 0);
                DIA:    in std_logic_vector(15 downto 0);
                DIB:    in std_logic_vector(15 downto 0);
                WEA:    in std_logic;
                WEB:    in std_logic;
                CLKA:   in std_logic;
                CLKB:   in std_logic;
                RSTA:   in std_logic;
                RSTB:   in std_logic;
                ENA:    in std_logic;
                ENB:    in std_logic;
                DOA:    out std_logic_vector(15 downto 0);
                DOB:    out std_logic_vector(15 downto 0)
	);
	end component;
	
	signal reset					: std_logic; -- reset for brams
	signal read_enable				: std_logic; -- read_enable allows reading from brams 
	signal blank_enable				: std_logic; -- blank enable disallows writin to b-port of every bram
	signal enable_brams_a, enable_brams_b 		: std_logic; -- signals for enabling brams (doesn't really work)
	signal counter					: std_logic_vector(7 downto 0); -- counter of elements in bram
	signal temp_counter				: std_logic_vector(3 downto 0); -- used in reading from bram/writing to smp
	signal size					: std_logic_vector(2 downto 0); -- used in reading from bram/writing to smp
	signal read_address, write_address 		: std_logic_vector(7 downto 0); -- addresses for brams
	signal r1, r2, r3, r4, w1, w2, w3, w4		: std_logic_vector(15 downto 0); -- signals for brams
	signal do1, do2, do3, do4, do5, do6, do7, do8 	: std_logic_vector(15 downto 0); -- signals for brams
	signal addr_ptr					: std_logic_vector(39 downto 3); -- address for writing to smp
	signal valid_bram				: std_logic; -- valid signal when writing data to bram
	
begin

	write_main : process (reset_n, user_clk)
	variable inc_wa, inc_ra : std_logic;
	begin
		reset <= not reset_n; -- Modelsim won't allow negators on bram port map
		if reset_n='0' then
			ureq_addr     <= (others => '0');
			ureq_size     <= (others => '0');
			ureq_mask     <= (others => '1');
			ureq_rw_n     <= '0';
			ureq_data     <= (others => '0');
			ureq_ts       <= '0';
			ureq_byte_req <= '0';

			write_address <= (others => '1');
			read_address <= (others => '0');
			counter <= (others => '0');
			temp_counter <= (others => '0');
			read_enable <= '0';
			enable_brams_a <= '0';
			enable_brams_b <= '0';
			blank_enable <= '0';
			
			addr_ptr <= (others => '0');

			stall <= '0';
			overflow <= '0';
			empty <= '1';
		
		elsif user_clk='1' and user_clk'event then
			-- Initiation
			enable_brams_a <= '1';
			blank_enable <= '0';
			inc_ra := '0';
			inc_wa := '0';

			-- Set the different state-signals
			overflow <= and_reduce(counter&valid);
			stall <= and_reduce(counter) or and_reduce(counter+valid) or ureq_full;
			empty <= not or_reduce(counter&valid);
			enable_brams_b <= or_reduce(counter&valid);

			-- write to brams
			if valid='1' then
				w1 <= data_in(63 downto 48); -- Distribute data to different brams
				w2 <= data_in(47 downto 32); 
				w3 <= data_in(31 downto 16);
				w4 <= data_in( 15 downto 0);
				inc_wa := '1'; -- Increase write_address
				valid_bram <= '1';
			else
				valid_bram <= '0';
			end if;

			-- read from bram and write to smp
			if temp_counter = 8 then -- Burstwrite startup
				ureq_ts <= '1'; -- Assign signal for burstwrite startup
				ureq_rw_n <= '0'; -----||-----
				ureq_byte_req <= '0'; ----||-----
				ureq_addr <= addr_ptr; ------||------		
				ureq_size <= not size;
				temp_counter <= (B"1" & size) + 1;
				ureq_data <= do2 & do4 & do6 & do8; -- Only even numbers because we are reading port b of every bram
				inc_ra := or_reduce(size+1);
				addr_ptr <= addr_ptr + 1;
			elsif temp_counter = 0 then -- No ongoing transfer to smp, possibly except for cooldown-phase
				if counter > 7 then -- If the memory is ready to transfer (no matter what flush is like)
					temp_counter <= B"1000";
					size <= B"000"; -- set size to 0 (which actually is 7)
					inc_ra := '1';
				elsif	(flush = '1' and counter(2 downto 0) > 0) then -- If flush = 1 and 8>counter>0...
					temp_counter <= B"1000"; -- Initiate transfer
					size <= not counter(2 downto 0) + 1; -- set size to 1 and nr of elements in bram
					inc_ra := '1';
				elsif set_base_addr = '1' then -- If new addr_ptr ...
					addr_ptr <= addr_in;
				end if;
				ureq_ts <= '0'; -- Acts as cooldown fase in addition to setting temp_counter for next cycle
				ureq_addr <= (others => '0');
				ureq_size <= (others => '0');
				ureq_data <= (others => '0');
			else -- Push data
				ureq_data <= do2 & do4 & do6 & do8; -- Only even numbers because we are reading port b of every bram
				if temp_counter(2 downto 0) < 7 then -- If not the last cycle of the transfer, increment read_address
					inc_ra := '1';
				end if;
				addr_ptr <= addr_ptr + 1; -- Increment addr_ptr for every iteration of transfer to get correct next address
				temp_counter <= temp_counter + 1; -- Iterator
				ureq_ts <= '0';
				ureq_addr <= (others => '0');
				ureq_size <= (others => '0');
			end if;
				
			counter <= counter + inc_wa - inc_ra; -- Update counter and addresses
			read_address <= read_address + inc_ra;
			write_address <= write_address + inc_wa;
			
		end if;
	end process write_main;	

bram1 : RAMB4_S16_S16
	port map (
		ADDRA 	=> write_address,
		ADDRB 	=> read_address,
		DIA	=> w1,
		DIB	=> r1,
		WEA	=> valid_bram,
		WEB 	=> blank_enable, -- null
		CLKA	=> user_clk,
		CLKB	=> user_clk,
		RSTA	=> reset,
		RSTB	=> reset,
		ENA	=> enable_brams_a,
		ENB	=> enable_brams_b,
		DOA	=> do1,
		DOB	=> do2
	);
bram2 : RAMB4_S16_S16
	port map (
		ADDRA 	=> write_address,
		ADDRB 	=> read_address,
		DIA	=> w2,
		DIB	=> r2,
		WEA	=> valid_bram,
		WEB 	=> blank_enable, -- null
		CLKA	=> user_clk,
		CLKB	=> user_clk,
		RSTA	=> reset,
		RSTB	=> reset,
		ENA	=> enable_brams_a,
		ENB	=> enable_brams_b,
		DOA	=> do3,
		DOB	=> do4
	);
bram3 : RAMB4_S16_S16
	port map (
		ADDRA 	=> write_address,
		ADDRB 	=> read_address,
		DIA	=> w3,
		DIB	=> r3,
		WEA	=> valid_bram,
		WEB 	=> blank_enable, -- null
		CLKA	=> user_clk,
		CLKB	=> user_clk,
		RSTA	=> reset,
		RSTB	=> reset,
		ENA	=> enable_brams_a,
		ENB	=> enable_brams_b,
		DOA	=> do5,
		DOB	=> do6
	);
bram4 : RAMB4_S16_S16
	port map (
		ADDRA 	=> write_address,
		ADDRB 	=> read_address,
		DIA	=> w4,
		DIB	=> r4,
		WEA	=> valid_bram,
		WEB 	=> blank_enable, -- null
		CLKA	=> user_clk,
		CLKB	=> user_clk,
		RSTA	=> reset,
		RSTB	=> reset,
		ENA	=> enable_brams_a,
		ENB	=> enable_brams_b,
		DOA	=> do7,
		DOB	=> do8
	);
end rtl;

