--********************************************************************
--********************************************************************
--** This model is the property of Cypress Semiconductor Corp. and  **
--** is protected by the US copyright laws, any unauthorized copying** 
--** and distribution is prohibited. Cypress reserves the right to  **
--** change any of the functional specifications without any prior  **
--** notice. Cypress is not liable for any damages which may result **
--** from the use of this functional model.                         ** 
--**                                                                **
--** File Name: CY7C1314v18.vhd                                     **
--**                                                                **
--** Revision : 1.0                                                 **
--**                                                                **
--** All the timings to be assigned by the user depending on the    **
--** frequency of operation.                                        **
--**                                                                ** 
--** Model    : 512Kx36 QDRII burst-of-2                                        **
--**                                                                **
--** Queries?: www.cypress.com/support                              **
--********************************************************************
--********************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
use IEEE.STD_LOGIC_ARITH.all;

entity CY7C1314v18 is
  generic (
    -- Constant Parameters
    addr_bits : integer := 18;          -- This is external address
    data_bits : integer := 36;
    mem_sizes : integer := 524287;

    -- Clock Times for 200 Mhz
    tCYC : time := 5.00 ns;
    tKH  : time := 2.00 ns;
    tKL  : time := 2.00 ns;

    -- Output Times
    tCO : time := 0.38 ns;
--    tCO : time := 2.5 ns;

    -- Setup Times
    tSA : time := 0.50 ns;
    tSC : time := 0.50 ns;
    tSD : time := 0.50 ns;

    -- Hold Times
    tHA : time := 0.50 ns;
    tHC : time := 0.50 ns;
    tHD : time := 0.50 ns
    );

  port (
    D     : in  std_logic_vector (data_bits - 1 downto 0) := (others => 'Z');
    Q     : out std_logic_vector (data_bits - 1 downto 0) := (others => 'Z');
    A     : in  std_logic_vector (addr_bits - 1 downto 0);
    RPS_n : in  std_logic;
    WPS_n : in  std_logic;
    BWS_n : in  std_logic_vector (3 downto 0);
    K     : in  std_logic;
    K_n   : in  std_logic;
    C     : in  std_logic;
    C_n   : in  std_logic;
    CQ    : out std_logic;
    CQ_n  : out std_logic
    );
end CY7C1314v18;

architecture behave of CY7C1314v18 is
  type memory is array (mem_sizes downto 0) of std_logic_vector (data_bits - 1 downto 0);
  type a4xdata is array (3 downto 0) of std_logic_vector (data_bits - 1 downto 0);
  type a4xaddr is array (3 downto 0) of std_logic_vector (addr_bits - 1 downto 0);
  type a4xcmnd is array (3 downto 0) of std_logic;

  signal C_Int, C_Int_n, C_chk  : std_logic;
  signal Output_Buf, Output_reg : std_logic_vector (data_bits - 1 downto 0) := (others => 'Z');
begin
  -- Output Buffer
  Q <= Output_buf;

  -- Internal Output Clock Generators
  C_Int   <= transport K   after tCO when (C_chk = '1') else C after tCO;
  C_Int_n <= transport K_n after tCO when (C_chk = '1') else C_n after tCO;

  -- Echo Clock Generator
  CQ   <= C_Int;
  CQ_n <= C_Int_n;

  main : process
    -- Memory Array
    variable Bank0 : memory;

    -- Declar Connection Variables
    variable Data_in_enable  : std_logic := '0';
    variable Data_out_enable : std_logic := '0';

    variable Addr_read_reg : std_logic_vector (addr_bits - 1 downto 0);

    variable Bw_reg    : std_logic_vector (3 downto 0);
    variable Input_buf : std_logic_vector (data_bits - 1 downto 0);
    variable Input_reg : std_logic_vector (data_bits - 1 downto 0);

    -- Pipeline Variables
    variable Data_pipe : a4xdata;
    variable Addr_pipe : a4xaddr;
    variable Cmnd_pipe : a4xcmnd;

    -- Counter
    variable Burst_read_counter : std_logic;

  begin
    wait on K, K_n, C_Int, C_Int_n;

    -- Data Out
    if ((K'event and K = '1') or (K_n'event and K_n = '1')) then
      -- Check C toggle
      if C = '1' and C_n = '1' then
        C_chk <= '1';
      else
        C_chk <= '0';
      end if;

      -- Command Pipeline
      Cmnd_pipe (0) := Cmnd_pipe (1);
      Cmnd_pipe (1) := Cmnd_pipe (2);
      Cmnd_pipe (2) := Cmnd_pipe (3);
      Cmnd_pipe (3) := 'X';

      -- Address Pipeline
      Addr_pipe (0) := Addr_pipe (1);
      Addr_pipe (1) := Addr_pipe (2);
      Addr_pipe (2) := Addr_pipe (3);
      Addr_pipe (3) := (others => '0');

      -- Data Pipeline
      Data_pipe (0) := Data_pipe (1);
      Data_pipe (1) := Data_pipe (2);
      Data_pipe (2) := Data_pipe (3);
      Data_pipe (3) := (others => 'X');

      -- Internal Read Command
      if (Cmnd_pipe (0) = '1') then
        Data_out_enable    := '1';
        Addr_read_reg      := Addr_pipe (0);
        Burst_read_counter := '0';
      end if;

      -- Data Out Register
      if (Data_out_enable = '1') then
        -- Read Data Into Output Register
        Output_reg <= Data_pipe(0) after tCO;

        -- Reset
        if (Burst_read_counter = '1' and Data_out_enable = '1') then
          Data_out_enable := '0';
        end if;

        -- Increasement Counter
        Burst_read_counter := not (Burst_read_counter);

      else
        Output_reg <= (others => 'Z') after tCO;
      end if;
    end if;

    -- Latch External Command
    if (K'event and K = '1') then
      if (RPS_n = '0') then
        Cmnd_pipe (3) := '1';
        Addr_pipe (3) := A;
      end if;
      if (WPS_n = '0') then
        Data_in_enable := '1';
        Input_buf      := D;
        Bw_reg         := BWS_n;
      end if;
    end if;

    -- Data In
    if (K_n'event and K_n = '1') then
      if (Data_in_enable = '1') then
        -- First Write Data (latch data from posedge K)
        if (Bw_reg (0) = '0' or Bw_reg (1) = '0' or Bw_reg (2) = '0' or Bw_reg (3) = '0') then
          -- Read Data Into Input Register
          Input_reg := Bank0 (CONV_INTEGER(A & '0'));

          -- Perform BW# Operation
          if (Bw_reg (0) = '0') then
            Input_reg (8 downto 0) := Input_buf (8 downto 0);
          end if;
          if (Bw_reg (1) = '0') then
            Input_reg (17 downto 9) := Input_buf (17 downto 9);
          end if;
          if (Bw_reg (2) = '0') then
            Input_reg (26 downto 18) := Input_buf (26 downto 18);
          end if;
          if (Bw_reg (3) = '0') then
            Input_reg (35 downto 27) := Input_buf (35 downto 27);
          end if;

          -- Write Back to Memory
          Bank0 (CONV_INTEGER(A & '0')) := Input_reg;

          -- Compare address to see if data forwarded or buffered from memory array
          if Addr_pipe (2) = A then
            Data_pipe (2) := Input_reg;
          else
            Data_pipe (2) := Bank0 (CONV_INTEGER(Addr_pipe (2) & '0'));
          end if;
        else
          Data_pipe (2) := Bank0 (CONV_INTEGER(Addr_pipe (2) & '0'));
        end if;

        -- Second Write Data
        if (BWS_n (0) = '0' or BWS_n (1) = '0' or BWS_n (2) = '0' or BWS_n (3) = '0') then
          -- Read Data Into Input Register
          Input_reg := Bank0 (CONV_INTEGER(A & '1'));

          -- Perform BW# Operation
          if (BWS_n (0) = '0') then
            Input_reg (8 downto 0) := D (8 downto 0);
          end if;
          if (BWS_n (1) = '0') then
            Input_reg (17 downto 9) := D (17 downto 9);
          end if;
          if (BWS_n (2) = '0') then
            Input_reg (26 downto 18) := D (26 downto 18);
          end if;
          if (BWS_n (3) = '0') then
            Input_reg (35 downto 27) := D (35 downto 27);
          end if;

          -- Write Back to Memory
          Bank0 (CONV_INTEGER(A & '1')) := Input_reg;

          -- Compare address to see if data forwarded or buffered from memory array
          if Addr_pipe (2) = A then
            Data_pipe (3) := Input_reg;
          else
            Data_pipe (3) := Bank0 (CONV_INTEGER(Addr_pipe (2) & '1'));
          end if;
        else
          Data_pipe (3) := Bank0 (CONV_INTEGER(Addr_pipe (2) & '1'));
        end if;

        -- Reset
        Data_in_enable := '0';
      else
        Data_pipe (2) := Bank0 (CONV_INTEGER(Addr_pipe (2) & '0'));
        Data_pipe (3) := Bank0 (CONV_INTEGER(Addr_pipe (2) & '1'));
      end if;
    end if;

    --Output Buffer
    if ((C_Int'event and C_Int = '1') or (C_Int_n'event and C_Int_n = '1')) then
      Output_buf <= Output_reg;
    end if;
  end process main;

  -- Check for K Timing Violation
  K_check : process
    variable k_high, k_low : time := 0 ns;
  begin
    wait on K;
    if K = '1' and NOW >= tCYC then
      assert (NOW - k_low >= tKH)
        report "K width low - tKH violation"
        severity error;
      assert (NOW - k_high >= tCYC)
        report "K period high - tCYC violation"
        severity error;
      k_high := NOW;
    elsif K = '0' and NOW >= tCYC then
      assert (NOW - k_high >= tKL)
        report "K width high - tKL violation"
        severity error;
      assert (NOW - k_low >= tCYC)
        report "K period low - tCYC violation"
        severity error;
      k_low := NOW;
    end if;
  end process;

  -- Check for K# Timing Violation
  K_n_check : process
    variable k_n_high, k_n_low : time := 0 ns;
  begin
    wait on K_n;
    if K_n = '1' and NOW >= tCYC then
      assert (NOW - k_n_low >= tKH)
        report "K# width low - tKH violation"
        severity error;
      assert (NOW - k_n_high >= tCYC)
        report "K# period high - tCYC violation"
        severity error;
      k_n_high := NOW;
    elsif K_n = '0' and NOW >= tCYC then
      assert (NOW - k_n_high >= tKL)
        report "K# width high - tKL violation"
        severity error;
      assert (NOW - k_n_low >= tCYC)
        report "K# period low - tCYC violation"
        severity error;
      k_n_low := NOW;
    end if;
  end process;

  -- Check for C Timing Violation
  C_check : process
    variable c_high, c_low : time := 0 ns;
  begin
    wait on C;
    if C = '1' and NOW >= tCYC then
      assert (NOW - c_low >= tKH)
        report "C width low - tKH violation"
        severity error;
      assert (NOW - c_high >= tCYC)
        report "C period high - tCYC violation"
        severity error;
      c_high := NOW;
    elsif C = '0' and NOW >= tCYC then
      assert (NOW - c_high >= tKL)
        report "C width high - tKL violation"
        severity error;
      assert (NOW - c_low >= tCYC)
        report "C period low - tCYC violation"
        severity error;
      c_low := NOW;
    end if;
  end process;

  -- Check for C# Timing Violation
  C_n_check : process
    variable c_n_high, c_n_low : time := 0 ns;
  begin
    wait on C_n;
    if C_n = '1' and NOW >= tCYC then
      assert (NOW - c_n_low >= tKH)
        report "C# width low - tKH violation"
        severity error;
      assert (NOW - c_n_high >= tCYC)
        report "C# period high - tCYC violation"
        severity error;
      c_n_high := NOW;
    elsif K_n = '0' and NOW >= tCYC then
      assert (NOW - c_n_high >= tKL)
        report "C# width high - tKL violation"
        severity error;
      assert (NOW - c_n_low >= tCYC)
        report "C# period low - tCYC violation"
        severity error;
      c_n_low := NOW;
    end if;
  end process;

  -- Check for Setup Timing Violation
  setup_check : process
  begin
    wait on K;
    if K = '1' then
      assert (RPS_n'last_event >= tSA)
        report "R# - tSC violation"
        severity error;
      assert (WPS_n'last_event >= tSA)
        report "W# - tSC violation"
        severity error;
      assert (A'last_event >= tSA)
        report "SA - tSA violation"
        severity error;
      assert (D'last_event >= tSA)
        report "D - tSD violation"
        severity error;
    end if;
  end process;

  -- Check for Hold Timing Violation
  hold_check : process
  begin
    wait on K'delayed(tHC), K'delayed(tHA), K'delayed(tHD);
    if K'delayed(tHC) = '1' then
      assert (RPS_n'last_event > tHC)
        report "R# - tHC violation"
        severity error;
      assert (WPS_n'last_event > tHC)
        report "W# - tHC violation"
        severity error;
    end if;
    if K'delayed(tHA) = '1' then
      assert (A'last_event > tHA)
        report "SA - tHA violation"
        severity error;
    end if;
    if K'delayed(tHD) = '1' then
      assert (D'last_event > tHD)
        report "D - tHD violation"
        severity error;
    end if;
  end process;
end behave;
