-------------------------------------------------------------------------------
-- Title      : Fabric Model
-- Project    : 
-------------------------------------------------------------------------------
-- File       : $RCSfile: fabric_model.vhd,v $
-- Author     : Cray Canada
-- Company    : Cray Canada Inc.
-- Created    : 2003-09-18
-- Last update: 2005-02-11
-------------------------------------------------------------------------------
-- Description: This file provides a simple VHDL model for the Rapid Array
-- Fabric.
-------------------------------------------------------------------------------
-- Copyright (c) 2004 Cray Canada Inc.
-------------------------------------------------------------------------------
-- Revisions  :
-- $Log: fabric_model.vhd,v $
-- Revision 1.2  2005/02/25 21:43:27 
-- Cleaned up indentation.
--
-- Revision 1.1  2005/01/04 20:54:11 
-- Initial checkin.
--
-- Revision 1.6  2004/06/09 22:45:35 
-- Updated rt_core.
--
-- Revision 1.2  2004/05/28 18:52:58 
-- Renamed signals and library.
--
-- Revision 1.1  2004/05/24 22:21:07 
-- Initial checkin.
--
-- Revision 1.3  2004/04/27 23:14:08 
-- Got rid of tx_delay and added enhanced error output.
--
-- Revision 1.2  2004/04/15 18:55:25 
-- Fixed parity generation.
--
-- Revision 1.1  2004/03/08 18:26:52 
-- Initial checkin.
--
-- Revision 1.1  2003/12/29 22:04:42 
-- Name change.
--
-- Revision 1.5  2003/12/21 00:43:06 
-- Updated to use fabric_pkg.
--
-- Revision 1.4  2003/12/21 00:25:34 
-- Offset the transmit clock and data.
--
-- Revision 1.3  2003/12/12 22:26:08 
-- Improved response checking.
--
-- Revision 1.2  2003/12/12 21:40:19 
-- Added line printing when read response didn't match expected.
--
-- Revision 1.1  2003/12/08 21:14:08 
-- Initial checkin.
--
-------------------------------------------------------------------------------

library ieee, std;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;
use ieee.std_logic_misc.all;
use std.textio.all;
use work.fabric_pkg.all;

entity fabric_model is

  generic (
    fabric_input  : string := "fabric.in";
    fabric_output : string := "fabric.out"); 

  port (
    reset_n     : in  std_logic;
    ht_rx_clk_p : in  std_logic;
    ht_rx_clk_n : in  std_logic;
    ht_rx_ctl_p : in  std_logic;
    ht_rx_ctl_n : in  std_logic;
    ht_rx_cad_p : in  std_logic_vector(15 downto 0);
    ht_rx_cad_n : in  std_logic_vector(15 downto 0);
    ht_rx_par_p : in  std_logic_vector(1 downto 0);
    ht_rx_par_n : in  std_logic_vector(1 downto 0);
    ht_tx_clk_p : out std_logic                     := '1';
    ht_tx_clk_n : out std_logic                     := '0';
    ht_tx_ctl_p : out std_logic                     := '0';
    ht_tx_ctl_n : out std_logic                     := '1';
    ht_tx_cad_p : out std_logic_vector(15 downto 0) := (others => '1');
    ht_tx_cad_n : out std_logic_vector(15 downto 0) := (others => '0');
    ht_tx_par_p : out std_logic_vector(1 downto 0)  := (others => '1');
    ht_tx_par_n : out std_logic_vector(1 downto 0)  := (others => '0')
    );

end entity fabric_model;

architecture behave of fabric_model is


  -----------------------------------------------------------------------------
  -- Signal Declarations
  -----------------------------------------------------------------------------

  file input_file  : text open read_mode is fabric_input;
  file output_file : text open write_mode is fabric_output;

  -- Fabric Request Signals
  signal clk      : std_logic          := '0';
  signal tx_clk   : std_logic          := '1';
  signal exp_buff : cmd_array(0 to 31) := (others => null_cmd);
  signal tag_rptr : integer            := 0;
  signal tag_wptr : integer            := 0;
  signal tag_buff : tag_array(0 to 31) := ("00000", "00001", "00010", "00011",
                                           "00100", "00101", "00110", "00111",
                                           "01000", "01001", "01010", "01011",
                                           "01100", "01101", "01110", "01111",
                                           "10000", "10001", "10010", "10011",
                                           "10100", "10101", "10110", "10111",
                                           "11000", "11001", "11010", "11011",
                                           "11100", "11101", "11110", "11111");

  -- User Request/Response Signals
  signal rt_rx_clk   : std_logic;
  signal rx_cad_qw   : std_logic_vector(63 downto 0);
  signal rx_ctl_qw   : std_logic_vector(3 downto 0);
  signal rx_cad_qw_r : std_logic_vector(63 downto 0);
  signal rx_ctl_qw_r : std_logic_vector(3 downto 0);
  signal qw_valid    : std_logic;
  signal ctl_samp    : std_logic;
  signal cad_samp    : std_logic_vector(15 downto 0);
  signal demux_state : std_logic_vector(1 downto 0);
  signal demux_count : integer;
  signal ram         : qw_array(0 to 16383) := (others => (others => '0'));
  signal fresp_queue : cmd_array(0 to 31)   := (others => null_cmd);
  signal fresp_rptr  : integer;
  signal fresp_wptr  : integer;
  signal tx_cad_p    : std_logic_vector(15 downto 0);
  signal tx_cad_n    : std_logic_vector(15 downto 0);
  signal tx_par_p    : std_logic_vector(1 downto 0);
  signal tx_par_n    : std_logic_vector(1 downto 0);
  signal tx_ctl_p    : std_logic;
  signal tx_ctl_n    : std_logic;
  signal init        : std_logic;
  signal rx_ready    : std_logic;

  -- Statistics
  signal rx_cnt : stats;
  signal tx_cnt : stats;
  
begin  -- architecture behave

  -----------------------------------------------------------------------------
  -- Transmit processing
  -- This block generates Fabric requests and Fabric responses. Fabric
  -- requests are created based on the input text file. Fabric responses are
  -- generated when a user read request has been received.
  -----------------------------------------------------------------------------
  clk <= not clk after 0.625 ns;

  tx_clocks : process (clk, reset_n) is
  begin  -- process clocks
    if (clk'event and clk = '1') then
      tx_clk <= not tx_clk;
    end if;
  end process tx_clocks;

  ht_tx_clk_p <= tx_clk;
  ht_tx_clk_n <= not tx_clk;

  fabric_request : process (clk, reset_n) is
    variable tx_cmd      : cmd_line := null_cmd;
    variable file_done   : boolean  := false;
    variable cmd_done    : boolean  := false;
    variable cmd_phase   : integer  := 0;
    variable cmd_count   : integer  := 0;
    variable line_number : integer  := 0;
    variable delay_cnt   : integer  := 0;
    variable free_buf    : boolean  := false;
    variable stall       : boolean  := false;
    variable tx          : tx_bus;

  begin
    if reset_n = '0' then               -- asynchronous reset (active low)
      cmd_done   := false;
      cmd_phase  := 0;
      cmd_count  := 0;
      tx_ctl_p   <= '0';
      tx_ctl_n   <= '1';
      tx_cad_p   <= (others => '1');
      tx_cad_n   <= (others => '0');
      tx_par_p   <= (others => '1');
      tx_par_n   <= (others => '0');
      fresp_rptr <= 0;
      tx_cnt     <= (rd => 0, wr => 0, resp => 0);
      init       <= '1';
      tag_rptr   <= 0;
    elsif clk'event and clk = '1' then  -- rising clock edge
      if rx_ready = '0' then
        if not cmd_done and init = '1'then
          tx_cmd.letter := 'I';
          init          <= '0';
        elsif not cmd_done then
          cmd_phase := cmd_phase + 1;
        end if;
      end if;

      -- Fetch a command from the input text file or process a user read
      -- request. User requests have priority.
      if rx_ready = '1' then
        if cmd_done then
          cmd_phase := 0;
          cmd_done  := false;
          if (delay_cnt > 0) then
            delay_cnt := delay_cnt - 1;
          end if;
          if ((cmd_count mod 8) = 5) then
            tx_cmd.letter := '-';
            free_buf      := true;
          elsif (fresp_wptr /= fresp_rptr) then  -- process user read request
            tx_cmd := fresp_queue(fresp_rptr);
            if (fresp_rptr = 31) then
              fresp_rptr <= 0;
            else
              fresp_rptr <= fresp_rptr + 1;
            end if;
          elsif (delay_cnt > 0) then
            tx_cmd.letter := '-';
          elsif (not stall) then                 -- get a new comand from file
            get_command(input_file, fabric_input, rx_cnt, tx_cnt,
                        file_done, line_number, tx_cmd);
          end if;

          if (tx_cmd.letter = 'R') then  -- if file command is a read
            -- Grab the next source tag.
            stall := true;
            if (tag_rptr /= tag_wptr) then
              tx_cmd.srctag := tag_buff(tag_rptr);
              if (tag_rptr = 31) then
                tag_rptr <= 0;
              else
                tag_rptr <= tag_rptr + 1;
              end if;
              stall := false;
            end if;
            if (not stall) then
              exp_buff(to_integer(unsigned(tx_cmd.srctag))) <= tx_cmd;
            end if;
          end if;
          cmd_count := cmd_count + 1;
        else
          cmd_phase := cmd_phase + 1;
        end if;
      end if;
      -- Execute the appropriate command procedure.
      if (stall) then
        tx_nop(cmd_phase, tx_cmd, free_buf, tx , cmd_done);
      else
        case tx_cmd.letter is
          when 'R' =>                    -- Read request
            tx_read(cmd_phase, tx_cmd, tx_cnt.rd, tx , cmd_done);
          when 'W' =>                    -- Write request
            tx_write(cmd_phase, tx_cmd, tx_cnt.wr, tx , cmd_done);
          when 'D' =>                    -- Delay
            tx_nop(cmd_phase, tx_cmd, free_buf, tx , cmd_done);
            delay_cnt := tx_cmd.delay;
          when 'I' =>                    -- Init
            tx_init(cmd_phase, tx, cmd_done);
          when 'U' =>                    -- Respond to user request
            tx_response(cmd_phase, tx_cmd, tx_cnt.resp, tx, cmd_done);
          when '-' =>                    -- NOP
            tx_nop(cmd_phase, tx_cmd, free_buf, tx , cmd_done);
          when others =>
            tx_nop(cmd_phase, tx_cmd, free_buf, tx , cmd_done);
        end case;
      end if;
      tx_ctl_p <= tx.ctl_p;
      tx_ctl_n <= tx.ctl_n;
      tx_par_p <= tx.par_p;
      tx_par_n <= tx.par_n;
      tx_cad_p <= tx.cad_p;
      tx_cad_n <= tx.cad_n;
    end if;
  end process fabric_request;

  -- Drive the signals after a delay
  ht_tx_cad_p <= tx_cad_p after 625 ps;
  ht_tx_cad_n <= tx_cad_n after 625 ps;
  ht_tx_par_p <= tx_par_p after 625 ps;
  ht_tx_par_n <= tx_par_n after 625 ps;
  ht_tx_ctl_p <= tx_ctl_p after 625 ps;
  ht_tx_ctl_n <= tx_ctl_n after 625 ps;

  -----------------------------------------------------------------------------
  -- Receive processing.
  -- This block decodes user requests and validates user responses. User
  -- responses are compared against the expected data provided in the Fabric
  -- read request.
  -----------------------------------------------------------------------------
  rt_rx_clk <= ht_rx_clk_p after 0.625 ns;  -- 90 degree phase shift

  rx_demux : process (rt_rx_clk, reset_n) is
    variable seen_init : boolean := false;
  begin  -- process rx_demux
    if reset_n = '0' then               -- asynchronous reset (active low)
      rx_cad_qw   <= (others => '0');
      rx_cad_qw_r <= (others => '0');
      rx_ctl_qw   <= (others => '0');
      rx_ctl_qw_r <= (others => '0');
      qw_valid    <= '0';
      ctl_samp    <= '0';
      cad_samp    <= (others => '0');
      demux_state <= (others => '0');
      demux_count <= 0;
      rx_ready    <= '0';
    elsif rt_rx_clk'event and rt_rx_clk = '1' then  -- rising clock edge
      ctl_samp <= ht_rx_ctl_p;
      cad_samp <= ht_rx_cad_p;

      if (ctl_samp = '1' and cad_samp(0) = '1') then
        rx_ready <= '1';
      end if;

      if (demux_state /= "10" and ctl_samp = '0' and ht_rx_ctl_p = '1') then
        demux_state <= demux_state + 1;
      end if;

      if (demux_state = "10") then
        if (demux_count = 3) then
          demux_count <= 0;
        else
          demux_count <= demux_count + 1;
        end if;
      else
        demux_count <= 0;
      end if;

      rx_cad_qw(demux_count*16+15 downto demux_count*16) <= cad_samp;
      rx_ctl_qw(demux_count)                             <= ctl_samp;

      -- Drive valid flag
      if (demux_count = 1) then
        qw_valid <= '1';
      else
        qw_valid <= '0';
      end if;

    elsif rt_rx_clk'event and rt_rx_clk = '0' then  -- falling clock edge
      ctl_samp <= ht_rx_ctl_p;
      cad_samp <= ht_rx_cad_p;

      if (demux_state = "10") then
        if (demux_count = 3) then
          demux_count <= 0;
        else
          demux_count <= demux_count + 1;
        end if;
      else
        demux_count <= 0;
      end if;

      rx_cad_qw(demux_count*16+15 downto demux_count*16) <= cad_samp;
      rx_ctl_qw(demux_count)                             <= ctl_samp;

      -- Register the data
      if (demux_count = 0) then
        rx_cad_qw_r <= rx_cad_qw;
        rx_ctl_qw_r <= rx_ctl_qw;
      end if;

    end if;
  end process rx_demux;

  rx_decode : process (rt_rx_clk, reset_n) is
    variable rx_cmd    : cmd_line := null_cmd;
    variable cmd_phase : integer  := 0;

  begin  -- process rx_decode
    if reset_n = '0' then               -- asynchronous reset (active low)
      fresp_wptr <= 0;
      tag_wptr   <= 31;
      rx_cnt     <= (rd => 0, wr => 0, resp => 0);
    elsif rt_rx_clk'event and rt_rx_clk = '1' then  -- rising clock edge
      if (qw_valid = '1') then
        if (rx_ctl_qw_r = "1111") then  -- command word
          cmd_phase                := 0;
          rx_cmd.addr(39 downto 2) := slv_to_req(rx_cad_qw_r).addr;
          rx_cmd.size              := slv_to_req(rx_cad_qw_r).mask_cnt;
          rx_cmd.srctag            := slv_to_req(rx_cad_qw_r).srctag;
          case slv_to_req(rx_cad_qw_r).cmd is
            when rt_post_byte_wr =>
              rx_cmd.letter   := 'W';
              rx_cmd.byte_req := "0001";
            when rt_post_dw_wr =>
              rx_cmd.letter   := 'W';
              rx_cmd.byte_req := "0000";
            when rt_byte_rd =>
              rx_cmd.letter   := 'R';
              rx_cmd.byte_req := "0001";
            when rt_dw_rd =>
              rx_cmd.letter   := 'R';
              rx_cmd.byte_req := "0000";
            when rt_rd_resp =>
              rx_cmd.letter   := 'U';
              rx_cmd.byte_req := "0000";
            when rt_nop =>
              rx_cmd.letter   := '-';
              rx_cmd.byte_req := "0000";
            when rt_sync_cmd =>
              rx_cmd.letter   := '-';
              rx_cmd.byte_req := "0000";
            when others =>
              assert false report "Fabric model received unknown command."
                severity error;
          end case;
          if (rx_cmd.letter = 'R') then
            rx_read(rx_cmd, ram, fresp_wptr, fresp_queue);
            rx_cnt.rd <= rx_cnt.rd + 1;
          end if;
        else
          rx_cmd.data(cmd_phase) := rx_cad_qw_r;
          if (cmd_phase = to_integer(unsigned(rx_cmd.size))/2) then
            if rx_cmd.letter = 'U' then
              rx_response(rx_cmd, fabric_input, exp_buff, tag_wptr, tag_buff);
              rx_cnt.resp <= rx_cnt.resp + 1;
            end if;
            if rx_cmd.letter = 'W' then
              rx_write(rx_cmd, ram);
              rx_cnt.wr <= rx_cnt.wr + 1;
            end if;
          end if;
          cmd_phase := cmd_phase + 1;
        end if;
      end if;
    end if;
  end process rx_decode;
  
end architecture behave;
