-------------------------------------------------------------------------------
-- Title      : Fabric Model Package
-- Project    : 
-------------------------------------------------------------------------------
-- File       : $RCSfile: fabric_pkg.vhd,v $
-- Author     : Cray Canada
-- Company    : Cray Canada Inc.
-- Created    : 2003-12-20
-- Last update: 2004-12-01
-------------------------------------------------------------------------------
-- Description: This package file contains constants, types, procedures and
-- functions used by the fabric model (fabric_model.vhd).
-------------------------------------------------------------------------------
-- Copyright (c) 2004 Cray Canada Inc.
-------------------------------------------------------------------------------
-- Revisions  :
-- $Log: fabric_pkg.vhd,v $
-- Revision 1.1  2005/01/04 20:54:11 
-- Initial checkin.
--
-- Revision 1.3  2004/05/26 23:36:39 
-- Changed RT interface naming.
--
-- Revision 1.2  2004/04/15 18:54:53 
-- Updated command types.
--
-- Revision 1.1  2004/03/08 18:26:52 
-- Initial checkin.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_textio.all;
use ieee.std_logic_misc.all;
use std.textio.all;

package fabric_pkg is

-------------------------------------------------------------------------------
-- Declare Types
-------------------------------------------------------------------------------

  -- Create records of the control packets so that we can symbolically
  -- manipulate them. 
  type req_pkt is                              -- Define the request packet
  record
    seqid    : std_logic_vector(3 downto 0);   -- Sequence ID
    cmd      : std_logic_vector(5 downto 0);   -- Command
    passpw   : std_logic;                      -- Pass Posted writes
    unitid   : std_logic_vector(4 downto 0);   -- Unit ID number
    mask_cnt : std_logic_vector(3 downto 0);   -- Byte mask or data size
    compat   : std_logic;                      -- Subtractive decode
    srctag   : std_logic_vector(4 downto 0);   -- Source Tag (sequence num.)
    addr     : std_logic_vector(39 downto 2);  -- Address
  end record req_pkt;

  type rsp_pkt is                           -- Define the response packet
  record
    cmd    : std_logic_vector(5 downto 0);  -- Command
    passpw : std_logic;                     -- Pass posted writes
    isoc   : std_logic;                     -- Isochronous channels
    bridge : std_logic;                     -- Host Bridge
    unitid : std_logic_vector(4 downto 0);  -- Unit ID number
    count  : std_logic_vector(3 downto 0);  -- Data packet size
    srctag : std_logic_vector(4 downto 0);  -- Source Tag (sequence num.)
    rquid  : std_logic_vector(1 downto 0);  -- Request ID
    err    : std_logic_vector(1 downto 0);  -- Error code
  end record rsp_pkt;

  type nop_pkt is                                 -- Define the nop packet
  record
    cmd          : std_logic_vector(5 downto 0);  -- Command
    discon       : std_logic;                     -- Disconnect sequence
    postcmd      : std_logic_vector(1 downto 0);  -- Post command buffer
    postdata     : std_logic_vector(1 downto 0);  -- Post data buffer
    response     : std_logic_vector(1 downto 0);  -- Response command buffer
    responsedata : std_logic_vector(1 downto 0);  -- Response data buffer
    nonpostcmd   : std_logic_vector(1 downto 0);  -- Non post command buffer
    nonpostdata  : std_logic_vector(1 downto 0);  -- Non post data buffer
    isoc         : std_logic;                     -- Isochronous channels
    diag         : std_logic;                     -- Diagnostic (CRC)
  end record nop_pkt;

  type tx_bus is
  record
    cad_p : std_logic_vector(15 downto 0);
    cad_n : std_logic_vector(15 downto 0);
    par_p : std_logic_vector(1 downto 0);
    par_n : std_logic_vector(1 downto 0);
    ctl_p : std_logic;
    ctl_n : std_logic;
  end record tx_bus;

  type stats is
  record
    rd   : integer;
    wr   : integer;
    resp : integer;
  end record stats;

  type rt_cmd_type is (nop, byte_wr, dw_wr, byte_rd, dw_rd,
                       resp_rd, sync, unsupported);
  type mask_array is array (0 to 7) of std_logic_vector(7 downto 0);
  type dw_array is array (natural range <>) of std_logic_vector(31 downto 0);
  type qw_array is array (natural range <>) of std_logic_vector(63 downto 0);
  type block_array is array (0 to 15) of qw_array(0 to 7);
  type tag_array is array (natural range <>) of std_logic_vector(4 downto 0);

  -- Create a record to hold the fields read in from the text file.
  type cmd_line is record
    line_num : integer;
    letter   : character;
    delay    : integer;
    byte_req : std_logic_vector(3 downto 0);
    size     : std_logic_vector(3 downto 0);
    data     : qw_array(0 to 7);
    mask     : mask_array;
    addr     : std_logic_vector(39 downto 0);
    srctag   : std_logic_vector(4 downto 0);
  end record cmd_line;

  type cmd_array is array (natural range <>) of cmd_line;

-------------------------------------------------------------------------------
-- Declare Constants
-------------------------------------------------------------------------------

  -- Define the supported command codes.
  constant rt_nop          : std_logic_vector := "000000";
  constant rt_post_byte_wr : std_logic_vector := "101001";
  constant rt_post_dw_wr   : std_logic_vector := "101101";
  constant rt_byte_rd      : std_logic_vector := "010001";
  constant rt_dw_rd        : std_logic_vector := "010101";
  constant rt_rd_resp      : std_logic_vector := "110000";
  constant rt_sync_cmd     : std_logic_vector := "111111";

  -- Define some null packet constants for initializing the record types.
  constant null_req : req_pkt := (seqid    => (others => '0'),
                                  cmd      => (others => '0'),
                                  passpw   => '0',
                                  unitid   => (others => '0'),
                                  mask_cnt => (others => '0'),
                                  compat   => '0',
                                  srctag   => (others => '0'),
                                  addr     => (others => '0'));
  constant null_rsp : rsp_pkt := (cmd    => (others => '0'),
                                  passpw => '0',
                                  isoc   => '0',
                                  bridge => '0',
                                  unitid => (others => '0'),
                                  count  => (others => '0'),
                                  srctag => (others => '0'),
                                  rquid  => (others => '0'),
                                  err    => (others => '0'));
  constant null_nop : nop_pkt := (cmd          => (others => '0'),
                                  discon       => '0',
                                  postcmd      => (others => '0'),
                                  postdata     => (others => '0'),
                                  response     => (others => '0'),
                                  responsedata => (others => '0'),
                                  nonpostcmd   => (others => '0'),
                                  nonpostdata  => (others => '0'),
                                  isoc         => '0',
                                  diag         => '0');

  constant null_cmd : cmd_line := (line_num => 0,
                                   letter   => '-',
                                   delay    => 0,
                                   mask     => (others => (others => '0')),
                                   byte_req => (others => '0'),
                                   size     => (others => '0'),
                                   data     => (others => (others => '0')),
                                   addr     => (others => '0'),
                                   srctag   => (others => '0')
                                   );


-------------------------------------------------------------------------------
-- Declare Functions
-------------------------------------------------------------------------------

  -- The following functions convert the packet records to standard logic
  -- vectors and vice versa.
  function req_to_slv (
    pkt : req_pkt)
    return std_logic_vector;

  function slv_to_req (
    data : std_logic_vector(63 downto 0))
    return req_pkt;

  function rsp_to_slv (
    pkt : rsp_pkt)
    return std_logic_vector;
  
  function slv_to_rsp (
    data : std_logic_vector(31 downto 0))
    return rsp_pkt;

  function slv_to_nop (
    data : std_logic_vector(31 downto 0))
    return nop_pkt;

  function nop_to_slv (
    pkt : nop_pkt)
    return std_logic_vector;

  -- Fetch a command from the input stimulous file
  procedure get_command(file input_file       :       text;
                        constant fabric_input : in    string;
                        signal   rx_cnt       : in    stats;
                        signal   tx_cnt       : in    stats;
                        variable file_done    : inout boolean;
                        variable line_number  : inout integer;
                        variable cmd          : inout cmd_line);

  -- The following procedures drive the various request types onto the TX bus.
  procedure tx_read(variable phase  : in    integer;
                    variable tx_cmd : in    cmd_line;
                    signal   cnt    : inout integer;
                    variable tx     : out   tx_bus;
                    variable done   : out   boolean);

  procedure tx_write(variable phase  : in    integer;
                     variable tx_cmd : in    cmd_line;
                     signal   cnt    : inout integer;
                     variable tx     : out   tx_bus;
                     variable done   : out   boolean);

  procedure tx_nop(variable phase    : in    integer;
                   variable tx_cmd   : in    cmd_line;
                   variable free_buf : inout boolean;
                   variable tx       : out   tx_bus;
                   variable done     : out   boolean);

  procedure tx_init(variable phase : in  integer;
                    variable tx    : out tx_bus;
                    variable done  : out boolean);

  procedure tx_response(variable phase  : in    integer;
                        variable tx_cmd : in    cmd_line;
                        signal   cnt    : inout integer;
                        variable tx     : out   tx_bus;
                        variable done   : out   boolean);

  -- The following procedures extract commands from the RX bus.
  procedure rx_read(variable cmd : in    cmd_line;
                    signal ram   : in    qw_array;
                    signal wptr  : inout integer;
                    signal queue : out   cmd_array);

  procedure rx_write(variable cmd : in  cmd_line;
                     signal ram   : out qw_array);

  procedure rx_response(variable cmd          : in    cmd_line;
                        constant fabric_input : in    string;
                        signal   exp_buff     : in    cmd_array;
                        signal   wptr         : inout integer;
                        signal   tag_buff     : out   tag_array);

end package fabric_pkg;

package body fabric_pkg is

  -----------------------------------------------------------------------------
  -- Convert a request packet record type to a standard logic vector
  -----------------------------------------------------------------------------
  function req_to_slv (
    pkt : req_pkt)
    return std_logic_vector is
    variable result : std_logic_vector(63 downto 0);
  begin
    result := pkt.addr & pkt.mask_cnt & pkt.compat & pkt.srctag & pkt.passpw &
              pkt.seqid(1 downto 0) & pkt.unitid & pkt.seqid(3 downto 2) &
              pkt.cmd;
    return result;
  end function req_to_slv;

  -----------------------------------------------------------------------------
  -- Convert a standard logic vector to a request packet record type
  -----------------------------------------------------------------------------
  function slv_to_req (
    data : std_logic_vector(63 downto 0))
    return req_pkt is
    variable result : req_pkt;
  begin
    result.addr     := data(63 downto 26);
    result.mask_cnt := data(25 downto 22);
    result.compat   := data(21);
    result.srctag   := data(20 downto 16);
    result.passpw   := data(15);
    result.seqid    := data(7 downto 6) & data(14 downto 13);
    result.unitid   := data(12 downto 8);
    result.cmd      := data(5 downto 0);
    return result;
  end function slv_to_req;

  -----------------------------------------------------------------------------
  -- Convert a response packet record type to a standard logic vector
  -----------------------------------------------------------------------------
  function rsp_to_slv (
    pkt : rsp_pkt)
    return std_logic_vector is
    variable result : std_logic_vector(31 downto 0);
  begin
    result := pkt.rquid & pkt.err(1) & "000" & pkt.count & pkt.err(0) &
              pkt.srctag & pkt.passpw & pkt.bridge & '0' & pkt.unitid &
              pkt.isoc & '0' & pkt.cmd;
    return result;
  end function rsp_to_slv;

  -----------------------------------------------------------------------------
  -- Convert a standard logic vector to a response packet record type
  -----------------------------------------------------------------------------
  function slv_to_rsp (
    data : std_logic_vector(31 downto 0))
    return rsp_pkt is
    variable result : rsp_pkt;
  begin
    result.rquid  := data(31 downto 30);
    result.err(1) := data(29);
    result.count  := data(25 downto 22);
    result.err(0) := data(21);
    result.srctag := data(20 downto 16);
    result.passpw := data(15);
    result.bridge := data(14);
    result.unitid := data(12 downto 8);
    result.isoc   := data(7);
    result.cmd    := data(5 downto 0);
    return result;
  end function slv_to_rsp;

  -----------------------------------------------------------------------------
  -- Convert a NOP packet record type to a standard logic vector
  -----------------------------------------------------------------------------
  function nop_to_slv (
    pkt : nop_pkt)
    return std_logic_vector is
    variable result : std_logic_vector(31 downto 0);
  begin
    result := x"00" & '0' & pkt.diag & pkt.isoc & '0' & pkt.nonpostdata &
              pkt.nonpostcmd & pkt.responsedata & pkt.response &
              pkt.postdata & pkt.postcmd & '0' & pkt.discon & pkt.cmd;
    return result;
  end function nop_to_slv;

  -----------------------------------------------------------------------------
  -- Convert a standard logic vector to a NOP packet record type
  -----------------------------------------------------------------------------
  function slv_to_nop (
    data : std_logic_vector(31 downto 0))
    return nop_pkt is
    variable result : nop_pkt;
  begin
    result.diag         := data(22);
    result.isoc         := data(21);
    result.nonpostdata  := data(19 downto 18);
    result.nonpostcmd   := data(17 downto 16);
    result.responsedata := data(15 downto 14);
    result.response     := data(13 downto 12);
    result.postdata     := data(11 downto 10);
    result.postcmd      := data(9 downto 8);
    result.discon       := data(6);
    result.cmd          := data(5 downto 0);
    return result;
  end function slv_to_nop;

  -----------------------------------------------------------------------------
  -- Fetch the next command from the input text file.
  -----------------------------------------------------------------------------

  procedure get_command(file input_file       :       text;
                        constant fabric_input : in    string;
                        signal   rx_cnt       : in    stats;
                        signal   tx_cnt       : in    stats;
                        variable file_done    : inout boolean;
                        variable line_number  : inout integer;
                        variable cmd          : inout cmd_line) is
    variable lin         : line;
    variable lout        : line;
    variable is_hex      : boolean;
    variable is_int      : boolean;
    variable dropout     : boolean;
    variable burst_count : integer := 0;
    variable letter      : character;
    
  begin

    -- Defaults
    is_hex  := false;
    dropout := false;
    cmd     := null_cmd;

    -- Read through file until a valid command line is found then drop out
    -- from the loop.
    while (not file_done and not dropout) loop
      readline(input_file, lin);
      line_number := line_number + 1;
      if endfile(input_file)then        -- Check if we're at the end.
        file_done := true;
      end if;
      if (lin'length /= 0) then         -- There is text on the line
        -- Store the line number
        cmd.line_num := line_number;

        -- Read the command letter.
        read(lin, cmd.letter);

        -- If it's a read or write command ...
        if (cmd.letter = 'R' or cmd.letter = 'W') then
          -- Read the address.
          hread(lin, cmd.addr, is_hex);
          if is_hex = false then
            report fabric_input & ", line #" & integer'image(line_number) &
              ": Reading address field from file."
              severity error;
            cmd.addr := (others => '0');
          end if;

          -- Read the first data word.
          hread(lin, cmd.data(0), is_hex);
          if is_hex = false then
            report fabric_input & ", line #" & integer'image(line_number) &
              ": Reading data field from file."
              severity error;
            cmd.data(0) := (others => '0');
          end if;

          -- Read the data mask.
          hread(lin, cmd.mask(0), is_hex);
          if is_hex = false then
            report fabric_input & ", line #" & integer'image(line_number) &
              ": Reading byte mask field from file."
              severity error;
            cmd.mask(0) := (others => '0');
          end if;

          -- Read in the byte request flag.
          hread(lin, cmd.byte_req, is_hex);
          if is_hex = false then
            report fabric_input & ", line #" & integer'image(line_number) &
              ": Reading byte request field from file."
              severity error;
            cmd.byte_req := (others => '0');
          end if;

          -- Read in the size field.
          hread(lin, cmd.size, is_hex);
          if is_hex = false then
            report fabric_input & ", line #" & integer'image(line_number) &
              ": Reading size field from file."
              severity error;
            cmd.size := (others => '0');
          end if;

          -- If it is a burst command read the rest of the data.
          if (cmd.size > "0001") then
            burst_count        := 0;
            while (burst_count <= to_integer(unsigned(cmd.size))/2-1)
            loop
              readline(input_file, lin);
              line_number := line_number + 1;
              read(lin, letter);
              if (letter = 'B') then
                hread(lin, cmd.data(burst_count+1), is_hex);
                if is_hex = false then
                  report fabric_input & ", line #" &
                    integer'image(line_number) &
                    ": Reading burst data from file"
                    severity error;
                  cmd.data(burst_count+1) := (others => '0');
                end if;
                hread(lin, cmd.mask(burst_count+1), is_hex);
                if is_hex = false then
                  report fabric_input & ", line #" &
                    integer'image(line_number) &
                    ": Reading byte mask from file"
                    severity error;
                  cmd.mask(burst_count+1) := (others => '0');
                end if;
              else
                report fabric_input & ", line #" &
                  integer'image(line_number) &
                  ": Incomplete burst transaction. Missing some burst data."
                  severity error;
              end if;
              burst_count := burst_count + 1;
            end loop;
          end if;
          dropout := true;
        end if;

        -- If it's a Delay command read in the delay count.
        if (cmd.letter = 'D') then
          read(lin, cmd.delay, is_int);
          if is_int = false then
            report fabric_input & ", line #" & integer'image(line_number) &
              ": Reading delay from file"
              severity error;
            cmd.delay := 0;
          else
            cmd.delay := cmd.delay;
          end if;
          dropout := true;
        end if;

        -- If it is a init command drop out.
        if (cmd.letter = 'I') then
          dropout := true;
        end if;

        -- If it is a print command output the line.
        if (cmd.letter = 'P') then
          writeline(output, lin);
        end if;

        -- If its a stats command, dump the counters.
        if (cmd.letter = 'S') then
          write(lout,
                CR &
                "Fabric Model Statistics Report :" & CR &
                "Initiated: " & CR &
                "   Writes Issued: " & integer'image(tx_cnt.wr) & CR &
                "   Reads Issued: " & integer'image(tx_cnt.rd) & CR &
                "   Responses Obtained: " & integer'image(rx_cnt.resp) & CR &
                "Processed: " & CR &
                "   Writes Received: " & integer'image(rx_cnt.wr) & CR &
                "   Reads Received: " & integer'image(rx_cnt.rd) & CR &
                "   Responses Issued: " & integer'image(tx_cnt.resp) & CR);
          writeline(output, lout);
          assert tx_cnt.rd = rx_cnt.resp report
            "Not all reads received responses." severity warning;
        end if;

      end if;
    end loop;

  end get_command;

  -----------------------------------------------------------------------------
  -- Drive a Read Request onto the transmit bus
  -----------------------------------------------------------------------------
  procedure tx_read(variable phase  : in    integer;
                    variable tx_cmd : in    cmd_line;
                    signal   cnt    : inout integer;
                    variable tx     : out   tx_bus;
                    variable done   : out   boolean) is
    variable req     : req_pkt := null_req;
    variable req_slv : std_logic_vector(63 downto 0);
    variable word    : std_logic_vector(15 downto 0);
  begin
    req.addr   := tx_cmd.addr(39 downto 2);
    req.srctag := tx_cmd.srctag;
    if (tx_cmd.byte_req(0) = '1') then
      req.cmd      := rt_byte_rd;
      req.mask_cnt := tx_cmd.mask(0)(3 downto 0);
    else
      req.cmd      := rt_dw_rd;
      req.mask_cnt := tx_cmd.size;
    end if;
    req_slv     := req_to_slv(req);
    word        := req_slv(15+16*phase downto 16*phase);
    tx.cad_p    := word;
    tx.cad_n    := not word;
    tx.par_p(0) := xnor_reduce(word(7 downto 0));
    tx.par_p(1) := xnor_reduce(word(15 downto 8));
    tx.par_n(0) := xor_reduce(word(7 downto 0));
    tx.par_n(1) := xor_reduce(word(15 downto 8));
    tx.ctl_p    := '1';
    tx.ctl_n    := '0';
    if (phase > 2) then
      done := true;
      cnt  <= cnt + 1;
    else
      done := false;
      cnt  <= cnt;
    end if;
  end tx_read;

  -----------------------------------------------------------------------------
  -- Drive a Write Request onto the transmit bus
  -----------------------------------------------------------------------------
  procedure tx_write(variable phase  : in    integer;
                     variable tx_cmd : in    cmd_line;
                     signal   cnt    : inout integer;
                     variable tx     : out   tx_bus;
                     variable done   : out   boolean) is
    variable req     : req_pkt := null_req;
    variable req_slv : std_logic_vector(63 downto 0);
    variable word    : std_logic_vector(15 downto 0);
  begin
    req.addr := tx_cmd.addr(39 downto 2);
    if (tx_cmd.byte_req(0) = '1') then
      req.cmd      := rt_post_byte_wr;
      req.mask_cnt := tx_cmd.size + 1;  -- Account for mask data
    else
      req.cmd      := rt_post_dw_wr;
      req.mask_cnt := tx_cmd.size;
    end if;
    req_slv := req_to_slv(req);
    case phase is
      when 0|1|2|3 =>
        word        := req_slv(15+16*phase downto 16*phase);
        tx.cad_p    := word;
        tx.cad_n    := not word;
        tx.par_p(0) := xnor_reduce(word(7 downto 0));
        tx.par_p(1) := xnor_reduce(word(15 downto 8));
        tx.par_n(0) := xor_reduce(word(7 downto 0));
        tx.par_n(1) := xor_reduce(word(15 downto 8));
        tx.ctl_p    := '1';
        tx.ctl_n    := '0';
      when 4|5 =>
        if (tx_cmd.byte_req(0) = '0') then
          word := tx_cmd.data((phase-4)/4)
                  (15+16*(phase mod 4) downto 16*(phase mod 4));
          tx.cad_p    := word;
          tx.cad_n    := not word;
          tx.par_p(0) := xnor_reduce(word(7 downto 0));
          tx.par_p(1) := xnor_reduce(word(15 downto 8));
          tx.par_n(0) := xor_reduce(word(7 downto 0));
          tx.par_n(1) := xor_reduce(word(15 downto 8));
        else
          word := tx_cmd.mask(2*phase-7) &
                  tx_cmd.mask(2*phase-8);
          tx.cad_p    := word;
          tx.cad_n    := not word;
          tx.par_p(0) := xnor_reduce(word(7 downto 0));
          tx.par_p(1) := xnor_reduce(word(15 downto 8));
          tx.par_n(0) := xor_reduce(word(7 downto 0));
          tx.par_n(1) := xor_reduce(word(15 downto 8));
        end if;
        tx.ctl_p := '0';
        tx.ctl_n := '1';
      when others =>
        if (tx_cmd.byte_req(0) = '0') then
          word := tx_cmd.data((phase-4)/4)
                  (15+16*(phase mod 4) downto 16*(phase mod 4));
          tx.cad_p    := word;
          tx.cad_n    := not word;
          tx.par_p(0) := xnor_reduce(word(7 downto 0));
          tx.par_p(1) := xnor_reduce(word(15 downto 8));
          tx.par_n(0) := xor_reduce(word(7 downto 0));
          tx.par_n(1) := xor_reduce(word(15 downto 8));
          if (phase >= (7 + to_integer(unsigned(tx_cmd.size))/2*4)) then
            done := true;
            cnt  <= cnt + 1;
          else
            done := false;
            cnt  <= cnt;
          end if;
        else
          word := tx_cmd.data((phase-6)/4)
                  (15+16*((phase-2)mod 4) downto 16*((phase-2) mod 4));
          tx.cad_p    := word;
          tx.cad_n    := not word;
          tx.par_p(0) := xnor_reduce(word(7 downto 0));
          tx.par_p(1) := xnor_reduce(word(15 downto 8));
          tx.par_n(0) := xor_reduce(word(7 downto 0));
          tx.par_n(1) := xor_reduce(word(15 downto 8));
          if (phase >= (7 + ((to_integer(unsigned(tx_cmd.size)))+1)/2*4))
          then
            done := true;
            cnt  <= cnt + 1;
          else
            done := false;
            cnt  <= cnt;
          end if;
        end if;
        tx.ctl_p := '0';
        tx.ctl_n := '1';
    end case;
  end tx_write;

  -----------------------------------------------------------------------------
  -- Drive a NOP onto the transmit bus
  -----------------------------------------------------------------------------
  procedure tx_nop(variable phase    : in    integer;
                   variable tx_cmd   : in    cmd_line;
                   variable free_buf : inout boolean;
                   variable tx       : out   tx_bus;
                   variable done     : out   boolean) is
    variable curr_nop : nop_pkt                       := null_nop;
    variable nop_slv  : std_logic_vector(63 downto 0) := (others => '0');
    variable word     : std_logic_vector(15 downto 0) := (others => '0');
    
  begin
    -- Send full buffer credits in every NOP for now ...
    curr_nop.postcmd      := "11";
    curr_nop.postdata     := "11";
    curr_nop.response     := "11";
    curr_nop.responsedata := "11";
    curr_nop.nonpostcmd   := "11";
    curr_nop.nonpostdata  := "11";

    -- Send a NOP
    nop_slv := nop_to_slv(curr_nop) & nop_to_slv(curr_nop);
    if (free_buf) then
      word := nop_slv(15+16*phase downto 16*phase);
    else
      word := (others => '0');
    end if;
    tx.cad_p    := word;
    tx.cad_n    := not word;
    tx.par_p(0) := xnor_reduce(word(7 downto 0));
    tx.par_p(1) := xnor_reduce(word(15 downto 8));
    tx.par_n(0) := xor_reduce(word(7 downto 0));
    tx.par_n(1) := xor_reduce(word(15 downto 8));
    tx.ctl_p    := '1';
    tx.ctl_n    := '0';
    if (phase > 2) then
      done     := true;
      free_buf := false;
    else
      done     := false;
      free_buf := free_buf;
    end if;
  end tx_nop;

  -----------------------------------------------------------------------------
  -- Drive an Init Cycle onto the transmit bus
  -----------------------------------------------------------------------------

  procedure tx_init(variable phase : in  integer;
                    variable tx    : out tx_bus;
                    variable done  : out boolean) is
    variable word : std_logic_vector(15 downto 0);
  begin
    case phase is
      when 0 =>
        word        := (others => '1');
        tx.cad_p    := word;
        tx.cad_n    := not word;
        tx.par_p(0) := xnor_reduce(word(7 downto 0));
        tx.par_p(1) := xnor_reduce(word(15 downto 8));
        tx.par_n(0) := xor_reduce(word(7 downto 0));
        tx.par_n(1) := xor_reduce(word(15 downto 8));
        tx.ctl_p    := '1';
        tx.ctl_n    := '0';
      when 8 =>
        word        := (others => '0');
        tx.cad_p    := word;
        tx.cad_n    := not word;
        tx.par_p(0) := xnor_reduce(word(7 downto 0));
        tx.par_p(1) := xnor_reduce(word(15 downto 8));
        tx.par_n(0) := xor_reduce(word(7 downto 0));
        tx.par_n(1) := xor_reduce(word(15 downto 8));
        tx.ctl_p    := '0';
        tx.ctl_n    := '1';
      when 16 =>
        word        := (others => '1');
        tx.cad_p    := word;
        tx.cad_n    := not word;
        tx.par_p(0) := xnor_reduce(word(7 downto 0));
        tx.par_p(1) := xnor_reduce(word(15 downto 8));
        tx.par_n(0) := xor_reduce(word(7 downto 0));
        tx.par_n(1) := xor_reduce(word(15 downto 8));
        tx.ctl_p    := '0';
        tx.ctl_n    := '1';
      when 23 =>
        done := true;
      when others =>
        null;
    end case;
  end tx_init;

  -----------------------------------------------------------------------------
  -- Drive a Response onto the transmit bus
  -----------------------------------------------------------------------------

  procedure tx_response(variable phase  : in    integer;
                        variable tx_cmd : in    cmd_line;
                        signal   cnt    : inout integer;
                        variable tx     : out   tx_bus;
                        variable done   : out   boolean) is
    variable rsp     : rsp_pkt := null_rsp;
    variable rsp_slv : std_logic_vector(31 downto 0);
    variable word    : std_logic_vector(15 downto 0);
  begin
    case phase is
      when 0 | 1 =>
        rsp.count   := tx_cmd.size;
        rsp.srctag  := tx_cmd.srctag;
        rsp.cmd     := rt_rd_resp;
        rsp_slv     := rsp_to_slv(rsp);
        word        := rsp_slv(15+16*phase downto 16*phase);
        tx.cad_p    := word;
        tx.cad_n    := not word;
        tx.par_p(0) := xnor_reduce(word(7 downto 0));
        tx.par_p(1) := xnor_reduce(word(15 downto 8));
        tx.par_n(0) := xor_reduce(word(7 downto 0));
        tx.par_n(1) := xor_reduce(word(15 downto 8));
        tx.ctl_p    := '1';
        tx.ctl_n    := '0';
      when 2 | 3 =>
        tx.cad_p    := (others => '0');
        tx.cad_n    := (others => '1');
        tx.par_p(0) := '1';
        tx.par_p(1) := '1';
        tx.par_n(0) := '0';
        tx.par_n(1) := '0';
        tx.ctl_p    := '1';
        tx.ctl_n    := '0';
      when others =>
        word := tx_cmd.data((phase-4)/4)
                (15+16*(phase mod 4) downto 16*(phase mod 4));
        tx.cad_p    := word;
        tx.cad_n    := not word;
        tx.par_p(0) := xnor_reduce(word(7 downto 0));
        tx.par_p(1) := xnor_reduce(word(15 downto 8));
        tx.par_n(0) := xor_reduce(word(7 downto 0));
        tx.par_n(1) := xor_reduce(word(15 downto 8));
        tx.ctl_p    := '0';
        tx.ctl_n    := '1';
        if (phase >= (7 + ((to_integer(unsigned(tx_cmd.size)))/2*4)))
        then
          done := true;
          cnt  <= cnt + 1;
        else
          done := false;
          cnt  <= cnt;
        end if;
    end case;
  end tx_response;

  -----------------------------------------------------------------------------
  -- Extract information from a read request.
  -----------------------------------------------------------------------------
  
  procedure rx_read(variable cmd : in    cmd_line;
                    signal ram   : in    qw_array;
                    signal wptr  : inout integer;
                    signal queue : out   cmd_array) is
    variable address : integer := 0;
  begin
    queue(wptr).letter <= 'U';
    queue(wptr).size   <= cmd.size;
    queue(wptr).srctag <= cmd.srctag;
    queue(wptr).data   <= (others => (others => '0'));
    address            := to_integer(unsigned(cmd.addr(16 downto 3)));

    for i in 0 to to_integer(unsigned(cmd.size))/2 loop
      queue(wptr).data(i) <= ram(address+i);
    end loop;  -- i

    if (wptr = 31) then
      wptr <= 0;
    else
      wptr <= wptr + 1;
    end if;
  end rx_read;

  -----------------------------------------------------------------------------
  -- Extract information from a write request.
  -----------------------------------------------------------------------------
  
  procedure rx_write(variable cmd : in  cmd_line;
                     signal ram   : out qw_array) is
    variable receive_data : dw_array(0 to 15);
    variable address      : integer := 0;
    variable mask         : std_logic_vector(31 downto 0);
  begin
    address := to_integer(unsigned(cmd.addr(16 downto 3)));
    -- Store data array in array of 32 bit double words for convenience.
    for i in 0 to 7 loop
      receive_data(2*i)   := cmd.data(i)(31 downto 0);
      receive_data(2*i+1) := cmd.data(i)(63 downto 32);
    end loop;
    -- Depending on whether it's a byte request or not, we have to unpack and
    -- handle the data differently.
    if (cmd.byte_req = "0000") then
      -- Block write so store all the data up to the size specified.
      for i in 0 to to_integer(unsigned(cmd.size)) loop
        if (i mod 2 = 0) then
          ram(address)(31 downto 0) <= receive_data(i);
        else
          ram(address)(63 downto 32) <= receive_data(i);
          address                    := address + 1;
        end if;
      end loop;
    else
      -- Byte write so store all the data up to the size specified according
      -- to the mask.
      mask := receive_data(0);
      for dw in 0 to to_integer(unsigned(cmd.size)-2) loop
        if (dw mod 2 = 0) then
          for byte in 0 to 3 loop
            if (mask(dw*4+byte) = '1') then
              ram(address)(8*byte+7 downto 8*byte) <=
                receive_data(dw+1)(8*byte+7 downto 8*byte);
            end if;
          end loop;
        else
          for byte in 0 to 3 loop
            if (mask(dw*4+byte) = '1') then
              ram(address)(8*byte+7+32 downto 8*byte+32) <=
                receive_data(dw+1)(8*byte+7 downto 8*byte);
            end if;
          end loop;
          address := address + 1;
        end if;
      end loop;
    end if;
  end rx_write;

  -----------------------------------------------------------------------------
  -- Extract information from a response.
  -----------------------------------------------------------------------------

  procedure rx_response(variable cmd          : in    cmd_line;
                        constant fabric_input : in    string;
                        signal   exp_buff     : in    cmd_array;
                        signal   wptr         : inout integer;
                        signal   tag_buff     : out   tag_array) is
    variable receive_data : dw_array(0 to 15);
    variable correct_data : dw_array(0 to 15);
    variable line_number  : integer;
    variable srctag       : integer;
    variable rec          : line := null;
    variable exp          : line := null;
  begin
    -- Store line number and source tag.
    srctag      := to_integer(unsigned(cmd.srctag));
    line_number := exp_buff(srctag).line_num;
    if (wptr = 31) then
      wptr <= 0;
    else
      wptr <= wptr + 1;
    end if;
    tag_buff(wptr) <= cmd.srctag;

    -- Store data array in array of 32 bit double words for convenience.
    for i in 0 to 7 loop
      receive_data(2*i)   := cmd.data(i)(31 downto 0);
      receive_data(2*i+1) := cmd.data(i)(63 downto 32);
      correct_data(2*i)   := exp_buff(srctag).data(i)(31 downto 0);
      correct_data(2*i+1) := exp_buff(srctag).data(i)(63 downto 32);
    end loop;

    -- Compare valid words.
    for i in 0 to to_integer(unsigned(cmd.size)) loop
      if (receive_data(i) /= correct_data(i)) then
        exp := null;
        rec := null;
        hwrite(exp, correct_data(i), right, 8);
        hwrite(rec, receive_data(i), right, 8);
        report "Fabric model received unexpected read response data." & CR &
          "** Check " & fabric_input & " file, line #" &
          integer'image(line_number) & ". " &
          "Expected: " & exp.all & ". " &
          "Received: " & rec.all & "."
          severity error;
      end if;
      if (i mod 2 = 1) then
        line_number := line_number + 1;
      end if;
    end loop;
  end rx_response;

end package body fabric_pkg;
