/* Original file made by Lars Krutaadal, edited by perandg.

*  File takes dna file, pwm file and optionally threshold as input parameters.
*  If no threshold is given summation is used.

*/

#define _XOPEN_SOURCE 600
#include "ufplib.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <argp.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <math.h>

// DEFINE THE FPGA WRITE TYPES (ADDRESS OR DATA VALUE)
#define TYPE_VAL  0x0UL
#define TYPE_ADDR 0x1UL

// DEFINE THE ADDRESS OFFSETS FOR THE MINCE FPGA REGISTERS
#define APP_ID_REG       0x00UL
#define APP_CFG_REG      0x08UL
#define APP_LATCH_REG    0x10UL

#define SCALE 24
// DECLARE A TYPE FOR A 64 BIT UNSIGNED INTEGER.
typedef unsigned long u_64;
//typedef int fixedPoint;

// DECODE AND PRINT OUT ANY DRIVER ERRORS.
int print_err (err_e e)
{
  switch (e) {
    case NOERR:
      printf("Success.\n");
      break;
    case FILEOPRERR:
      printf("File operation system call failed.\n");
      break;
    case INVALIDOP:
      printf("Invalid API operation requested.\n");
      break;
    case INVALIDVAL:
      printf("Invalid value passed to the API call.\n");
      break;
    case INVALIDARGS:
      printf("Invalid argument passed to the API call.\n");
      break;
    case INVALIDINP:
      printf("Invalid input given to the API call.\n");
      break;
    case DEVOPRERR:
      printf("FPGA device operation error.\n");
      break;
    case UNKNOWNERR:
      printf("Unknown error.\n");
      break;
    default:
      break;
    }
    return 0;
}

// LOAD CONTENT FROM FILE AND COPY TO MEMORY
char * openDNAFile(char *filename) {
	FILE * pFile;
	unsigned long size;
	size_t result;
	char *buffer;

	// Open file
	pFile = fopen(filename, "r");
	if (pFile == NULL) {fputs("File error", stderr); exit(1);}

	// Obtain file size
	fseek(pFile, 0, SEEK_END);
	size = ftell(pFile);
	rewind(pFile);

	// Allocate memory to contain the whole file
	buffer = (char*) malloc(sizeof(char)*size);
	if (buffer == NULL) {fputs("Memory error", stderr); exit(2);}

	// Copy the file into the buffer
	result = fread(buffer, 1, size, pFile);
	if (result != size) {fputs("Reading error", stderr); exit(3);}

	// close file
	fclose(pFile);

	return buffer;
}

int fpEncode(float x) {
	return x * 16777216; //roundf()
}

float fpDecode(long x) {
	return (float) ((x) / 16777216.0);
}

// ????????? to float
float fpDecodeI(int x) {
	return (float) ((x) / 16777216.0);
}

// CREATE PWM-VALUE
int calcPWMvalue(float val, float divisor) {
	float floatTemp;
	double pwmTemp;

	// Safety, because log(0) == error.
	if (val == 0.0) val = 0.01; // redundant because all number + 0.01 on load

	pwmTemp = (double) ((val / divisor) / 0.25);
	floatTemp = (float) log(pwmTemp); // Need to go via double because of log(double x)-function

	return fpEncode(floatTemp);
}

// CREATE PWM-MATRIX
int *createPWM(float *prepwm) {
	int i, j;
	int *pwm;

	pwm = (int *) malloc(sizeof(int)*8*4);
	for(i=0; i<8; i++) {
		double divisor = (prepwm[i*4+0]+prepwm[i*4+1]+prepwm[i*4+2]+prepwm[i*4+3]);
		for(j=0; j<4; j++) {
			pwm[i*4+j]= calcPWMvalue(prepwm[i*4+j], divisor);
		}
	}
	return pwm;
}

// MAIN
int main (int argc, char **argv) {
  int fp_id;
  err_e e;

  float threshold = 0;
  int filter = 0;
  char *filename = "";
	char *pwmName = "";
	char *dnaName = "";

  switch (argc) {
    case 3: // summation
    	dnaName = argv[1];
    	pwmName = argv[2];
    	filter = 1;
    	filename = "top.bin.ufp.sum";
    	printf("Execution with summation filter starting with dna %s and pwm %s.\n", dnaName, pwmName);
    	break;
    case 4: // threshold
    	dnaName = argv[1];
    	pwmName = argv[2];
    	threshold = atof(argv[3]);
    	filter = 2;
    	filename = "top.bin.ufp.thresh";
    	printf("Execution with threshold filter starting, with dna %s, pwm %s and threshold is %f\n", dnaName, pwmName, threshold);
    	break;
    default:
    	printf("Parameters: ./fpwm [file containing dna] [file containing PWM] [integer threshold], if threshold is not given summation filter will be assumed\n");
    	printf("Examples:\n [./fpwm file.dna file.pwm 8] executes file.dna and file.pwm with a threshold of 8\n [./fpwm file.dna file.pwm] executes file.dna and file.pwm with summation filter\n");
    	break;
	}

	if (filter > 0) {

		// Read dna and pwm from file
		FILE * pFile;
		unsigned long dnaSize;
		char *dna;
		float *prepwm;
		int * pwm;

		// Load DNA
		dna = openDNAFile( dnaName );
		dnaSize = strlen(dna);

		// Load PWM
		prepwm = (float *) malloc(sizeof(float)*8*4);
		pFile = fopen (pwmName, "r");
		int i, j, num;
		float val;
    for (i=0; i<8; i++) {
      for (j=0; j<4; j++) {
				num = fscanf (pFile, "%f", &val);
				prepwm[i*4+j] = val + 0.01;
			}
		}
		fclose(pFile);

		pwm = createPWM(prepwm);

		// Open file for saving results
		pFile = fopen("output-fpwm.txt", "w");
		if (pFile==NULL) {fputs("File error", stderr); exit(1);}

		// Start FPGA
		if((fp_id = fpga_open ("/dev/ufp0", O_RDWR|O_SYNC, &e)) > 0) {

			printf("\nLoading FPGA..\n");

			// Load FPGA
			if(fpga_load(fp_id, filename, &e) > 0) {

				printf("FPGA successfully loaded with status 0x%02X\n\n", fpga_status(fp_id, &e));

				// Map FPGA memory to a local address

				u_64 *memmap = (u_64*) fpga_memmap(fp_id, 1<<27, PROT_READ|PROT_WRITE, MAP_SHARED, 0, &e);

				// Allocate an FPGA transfer region

				u_64 *ftrmem;

				// Make sure the needed mem-area fits in memory page size exactly (because of register_ftrmem)
				unsigned int ftrSize = 0;
				if ((16777216*8) % getpagesize() != 0) {
					ftrSize = (((16777216*8) % getpagesize()) + 1 ) * getpagesize();
				} else {
					ftrSize = 16777216 * 8;
				}

				// Align memory
				int status = posix_memalign((void **)&ftrmem, getpagesize(), ftrSize);

				if (status != 0) {
					printf("Error memalign, errorcode %d\n", status);
				} else {

					// Register shared memory
					status = fpga_register_ftrmem(fp_id, ftrmem, ftrSize, &e);
					if (status != 0) {
						printf("Allocation of shared memory failed with status %d\n", status);
					} else {

						printf("Allocated an FPGA transfer region of %d at 0x%lX ", ftrSize, (u_64)ftrmem); print_err(e);
						unsigned long i = 0;
						int j = 0;
						long DNAiterator = 0;
						int run = 0;
						float sum = 0.0;
						int nrOfResults = 0;
						unsigned long allTransferred = 0;


						// Loop to transfer the whole dna, in several parts if needed
						while (DNAiterator < dnaSize) {

							int k = 8;
							unsigned long other = 0;
							unsigned long transferred_elem = 0;

							// Rewind data for new cycle so that the search is complete
							// (this is crude and does not, like the program, handle "NNNN" in input)
							if (DNAiterator > 7) DNAiterator = DNAiterator - 7;

							run++;
							printf("\nTransferring to sram.");
							u_64 qw = 0x0000000000000000UL;

							// Add on-the-fly conversion here

							// Iterate through DNA until 16MB is transferred or end of DNA has occured
							while ((transferred_elem < 16777215) && (DNAiterator < dnaSize)) {
								// If vector contains 8 elements (64 bit) then transfer it
								if (k == 0) {
									k = 8;
									u_64 addr = ((1<<25) | (transferred_elem-8));
									fpga_wrt_appif_val(fp_id, qw, addr, 0, &e); // write using single writes
									qw = 0;
								}
								// Convert "a" to 0, "c" to 1, "g" to 2 and "t" to 3
								// Also construct 64 bit vector with 8 elements
								switch(dna[DNAiterator]) {
									case ('a'):
										k--;
										qw = (qw<<8) + (0x00);
										transferred_elem++;
										break;
									case ('c'):
										k--;
										qw = (qw<<8) + (0x01);
										transferred_elem++;
										break;
									case ('g'):
										k--;
										qw = (qw<<8) + (0x02);
										transferred_elem++;
										break;
									case ('t'):
										k--;
										qw = (qw<<8) + (0x03);
										transferred_elem++;
										break;
									default:
										// The start of a crude scheme to account for unmapped regions of the DNA
										// But nothing more has been implemented yet
										other++;
										break;
								}
								DNAiterator++;
							}

							// Write the last values, as a quadword with the values in more signifacant plositions.
							if (k != 8) {
								u_64 addr = ((1<<25) | (transferred_elem-(8-(k))));
								while (k != 0) {
									k--;
									qw = qw<<8 | (0x00);
								}
								fpga_wrt_appif_val(fp_id, qw, addr, 0, &e); // write using single writes
							}


							// If end of DNA is encountered (transferred_elem == 0) dont execute
							if (transferred_elem > 0) {
								printf(" Done\n");

								fpga_mem_sync(fp_id, &e);

								fflush(NULL);

								u_64 read_ftrmem, read_count, read_threshold;

								// Write PWM to FPGA
								if (run == 1) {
									printf("Wrote  PWM:\n");
									fprintf(pFile,"Wrote  PWM:\n");
								}

								// Iterate through pwm-matrix
								for(i=0; i<8; i++) {
									for(j=0; j<4; j++) {
										int data = pwm[i*4+j];
										u_64 addr = 1<<24 | i<<10 | j<<3;
										fpga_wrt_appif_val(fp_id, data, addr, 0, &e);
										if (run == 1) {
											printf("0x%08X\t", data);
											fprintf(pFile, "0x%08X\t", data);
										}
									}
									if (run == 1) {
										printf("\n");
										fprintf(pFile, "\n");
									}
								}

								// Calculate remaining runs and project it to user
								int remainingRuns = ((dnaSize-DNAiterator) / 16777215) + run;

								if (DNAiterator < dnaSize) printf("This is run %d of %d runs (roughly predicted), due to DNA size.\n", run, remainingRuns);
								else if (run == 1) printf("This is the only run.\n");
								else printf("This is the last run of %d runs\n", run);

								// Write values for execution to FPGA:
								// First memoryaddress
								fpga_wrt_appif_val(fp_id, (u_64)ftrmem, 0x50UL, 1, &e); // NOTE: type = 1 meaning address not value
								if (run == 1) {printf("Wrote ftr_pointer=%lX: ", (u_64)ftrmem); print_err(e); fflush(NULL);}

								fpga_rd_appif_val(fp_id, &read_ftrmem, 0x50UL, &e);
								if (run == 1) {printf("Read ftr_pointer=%lX: ", read_ftrmem); print_err(e); fflush(NULL);}

						 // Added by perandg, write threshold value to FPGA
								if (filter == 2) {
									fpga_wrt_appif_val(fp_id, fpEncode(threshold), 0xA0UL, 0, &e);
									if (run == 1) {printf("Wrote threshold=%f: ", threshold); print_err(e); fflush(NULL);}
									// Read value back
									fpga_rd_appif_val(fp_id, &read_threshold, 0xA0UL, &e);
									if (run == 1) {printf("Read threshold=%f: ",fpDecode(read_threshold)); print_err(e); fflush(NULL);}
								}
						// End add

								// Last number of elements in SRAM, this start computation on FPGA
								fpga_wrt_appif_val(fp_id, transferred_elem, 0x78UL, 0, &e);
								if (run == 1 || transferred_elem != 16777215) {printf("Wrote count=%lu: ", transferred_elem); print_err(e); fflush(NULL);}

								fpga_rd_appif_val(fp_id, &read_count, 0x78UL, &e);
								if (run == 1 || transferred_elem != 16777215) {printf("Read count=%lX: ", read_count); print_err(e); fflush(NULL);}

								usleep(200000); // 200.000us is about twice of the expected worst case time use of the FPWM when processing 16MB of data

								//printf("FPGA said:\n");

								int nulldata = 0;
								// Read results from shared memory
								for(i=0; i<transferred_elem; i++) {
									long data = *(ftrmem+i);
									long index = data >> 38;
									float result = fpDecodeI(data & 0xFFFFFFFF); // Use bitmask to get the relevant bits

									if (data != 0) {
										// Format according to filter
										if (filter == 1) {
											fprintf(pFile,"Sum: %ld\t(0x%016lX)\n", data, data);
											sum = sum + fpDecode(data);
										} else if (filter == 2) {
											fprintf(pFile,"Index: %ld\tResult: %f\t(0x%016lX)\n", index, result, data);
											nrOfResults++;
										}
									} else {
										nulldata++;
									}
								}

								if (nulldata != 0) printf("Results equal to 0 have not been printed.\n");

								// If another round is required to finish computation then reset FPGA
								if (DNAiterator < dnaSize) {
									status = fpga_reset(fp_id, &e);
									printf("Reset: "); print_err(e);
									usleep(1);
									status = fpga_start(fp_id, &e);
									printf("Restart: "); print_err(e); fflush(NULL);
									usleep(1);
								}
								allTransferred = allTransferred + transferred_elem;
							} else { // If there were no elements to process
								printf("\nThe run was not completed, because there were no elements to process\n");
							}
						} // End while, the whole DNA file has been processed

						// Deregister shared memory
						status = fpga_dereg_ftrmem(fp_id, ftrmem, &e);
						printf("\nDeregistering memory, status %d, ",status); print_err(e); fflush(NULL);

						// Print final stats according to filter
						if (filter == 1) {
							fprintf(pFile,"Final sum: %f, sum/8: %f\n", sum, sum/8);
							printf("Final sum: %f, sum/8: %f\n", sum, sum/8);
						} else if (filter == 2) {
							fprintf(pFile,"Nr of results printed: %d\n", nrOfResults);
							printf("Nr of results printed: %d\n", nrOfResults);
						}
						fprintf(pFile, "Nr of transferred elements: %lu\n", allTransferred);
						printf("Nr of transferred elements: %lu\n", allTransferred);
					} // fpga_register_ftrmem() was successful;
				} // Mem_align() was successful;
				free(dna);
				free(pwm);
				free(prepwm);
				fclose(pFile);
				printf("Results from computation is saved to file output-fpwm.txt, and the file is closed.\n");
			} else {
				printf("Failed to load FPGA.\n");
			}

			// Close the FPGA device
			fpga_close (fp_id, &e);

		}	else {
			printf("Failed to open FPGA.\n");
		}
	} // No filter was loaded
  return 0;
}
